#! /usr/bin/env python
"""SearchSource: directory containing PDF files (based on GROBID)"""
from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfinterp import resolve1
from pdfminer.pdfparser import PDFParser

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.crossref
import colrev.ops.built_in.search_sources.pdf_backward_search as bws
import colrev.ops.search
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class PDFSearchSource(JsonSchemaMixin):
    """SearchSource for PDF directories (based on GROBID)"""

    # pylint: disable=too-many-instance-attributes

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "file"
    search_type = colrev.settings.SearchType.PDFS
    api_search_supported = True
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "PDF directory"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/pdfs_dir.md"
    )

    __doi_regex = re.compile(r"10\.\d{4,9}/[-._;/:A-Za-z0-9]*")

    def __init__(
        self, *, source_operation: colrev.operation.CheckOperation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.source_operation = source_operation

        if not source_operation.review_manager.in_ci_environment():
            self.pdf_preparation_operation = (
                source_operation.review_manager.get_pdf_prep_operation(
                    notify_state_transition_operation=False
                )
            )

        self.pdfs_path = source_operation.review_manager.path / Path(
            self.search_source.search_parameters["scope"]["path"]
        )
        self.review_manager = source_operation.review_manager

        self.subdir_pattern: re.Pattern = re.compile("")
        self.r_subdir_pattern: re.Pattern = re.compile("")
        if "subdir_pattern" in self.search_source.search_parameters.get("scope", {}):
            self.subdir_pattern = self.search_source.search_parameters["scope"][
                "subdir_pattern"
            ]
            source_operation.review_manager.logger.info(
                f"Activate subdir_pattern: {self.subdir_pattern}"
            )
            if "year" == self.subdir_pattern:
                self.r_subdir_pattern = re.compile("([1-3][0-9]{3})")
            if "volume_number" == self.subdir_pattern:
                self.r_subdir_pattern = re.compile("([0-9]{1,3})(_|/)([0-9]{1,2})")
            if "volume" == self.subdir_pattern:
                self.r_subdir_pattern = re.compile("([0-9]{1,4})")
        self.crossref_connector = (
            colrev.ops.built_in.search_sources.crossref.CrossrefSearchSource(
                source_operation=source_operation
            )
        )

    def __update_if_pdf_renamed(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        record_dict: dict,
        records: dict,
        search_source: Path,
    ) -> bool:
        updated = True
        not_updated = False

        c_rec_l = [
            r
            for r in records.values()
            if f"{search_source}/{record_dict['ID']}" in r["colrev_origin"]
        ]
        if len(c_rec_l) == 1:
            c_rec = c_rec_l.pop()
            if "colrev_pdf_id" in c_rec:
                cpid = c_rec["colrev_pdf_id"]
                pdf_fp = search_operation.review_manager.path / Path(
                    record_dict["file"]
                )
                pdf_path = pdf_fp.parents[0]
                potential_pdfs = pdf_path.glob("*.pdf")

                for potential_pdf in potential_pdfs:
                    cpid_potential_pdf = colrev.record.Record.get_colrev_pdf_id(
                        review_manager=search_operation.review_manager,
                        pdf_path=potential_pdf,
                    )

                    if cpid == cpid_potential_pdf:
                        record_dict["file"] = str(
                            potential_pdf.relative_to(
                                search_operation.review_manager.path
                            )
                        )
                        c_rec["file"] = str(
                            potential_pdf.relative_to(
                                search_operation.review_manager.path
                            )
                        )
                        return updated
        return not_updated

    def __remove_records_if_pdf_no_longer_exists(
        self, *, search_operation: colrev.ops.search.Search
    ) -> None:
        # search_operation.review_manager.logger.debug(
        #     "Checking for PDFs that no longer exist"
        # )

        if not self.search_source.filename.is_file():
            return

        with open(self.search_source.filename, encoding="utf8") as target_db:
            search_rd = search_operation.review_manager.dataset.load_records_dict(
                load_str=target_db.read()
            )

        records = search_operation.review_manager.dataset.load_records_dict()

        to_remove: typing.List[str] = []
        files_removed = []
        for record_dict in search_rd.values():
            x_pdf_path = search_operation.review_manager.path / Path(
                record_dict["file"]
            )
            if not x_pdf_path.is_file():
                if records:
                    updated = self.__update_if_pdf_renamed(
                        search_operation=search_operation,
                        record_dict=record_dict,
                        records=records,
                        search_source=self.search_source.filename,
                    )
                    if updated:
                        continue
                to_remove.append(
                    f"{self.search_source.filename.name}/{record_dict['ID']}"
                )
                files_removed.append(record_dict["file"])

        search_rd = {
            x["ID"]: x
            for x in search_rd.values()
            if (search_operation.review_manager.path / Path(x["file"])).is_file()
        }

        if len(search_rd.values()) != 0:
            search_operation.review_manager.dataset.save_records_dict_to_file(
                records=search_rd, save_path=self.search_source.filename
            )

        if records:
            for record_dict in records.values():
                for origin_to_remove in to_remove:
                    if origin_to_remove in record_dict["colrev_origin"]:
                        record_dict["colrev_origin"].remove(origin_to_remove)
            if to_remove:
                search_operation.review_manager.logger.info(
                    f" {colors.RED}Removed {len(to_remove)} records "
                    f"(PDFs no longer available){colors.END}"
                )
                print(" " + "\n ".join(files_removed))
            records = {k: v for k, v in records.items() if v["colrev_origin"]}
            search_operation.review_manager.dataset.save_records_dict(records=records)
            search_operation.review_manager.dataset.add_record_changes()

    def __update_fields_based_on_pdf_dirs(
        self, *, record_dict: dict, params: dict
    ) -> dict:
        if not self.subdir_pattern:
            return record_dict

        if "journal" in params["scope"]:
            record_dict["journal"] = params["scope"]["journal"]
            record_dict["ENTRYTYPE"] = "article"

        if "conference" in params["scope"]:
            record_dict["booktitle"] = params["scope"]["conference"]
            record_dict["ENTRYTYPE"] = "inproceedings"

        if self.subdir_pattern:
            # Note : no file access here (just parsing the patterns)
            # no absolute paths needed
            partial_path = Path(record_dict["file"]).parents[0]

            if "year" == self.subdir_pattern:
                # Note: for year-patterns, we allow subfolders
                # (eg., conference tracks)
                match = self.r_subdir_pattern.search(str(partial_path))
                if match is not None:
                    year = match.group(1)
                    record_dict["year"] = year

            elif "volume_number" == self.subdir_pattern:
                match = self.r_subdir_pattern.search(str(partial_path))
                if match is not None:
                    volume = match.group(1)
                    number = match.group(3)
                    record_dict["volume"] = volume
                    record_dict["number"] = number
                else:
                    # sometimes, journals switch...
                    r_subdir_pattern = re.compile("([0-9]{1,3})")
                    match = r_subdir_pattern.search(str(partial_path))
                    if match is not None:
                        volume = match.group(1)
                        record_dict["volume"] = volume

            elif "volume" == self.subdir_pattern:
                match = self.r_subdir_pattern.search(str(partial_path))
                if match is not None:
                    volume = match.group(1)
                    record_dict["volume"] = volume

        return record_dict

    # curl -v --form input=@./profit.pdf localhost:8070/api/processHeaderDocument
    # curl -v --form input=@./thefile.pdf -H "Accept: application/x-bibtex"
    # -d "consolidateHeader=0" localhost:8070/api/processHeaderDocument
    def __get_record_from_pdf_grobid(
        self, *, search_operation: colrev.ops.search.Search, record_dict: dict
    ) -> dict:
        if colrev.record.RecordState.md_prepared == record_dict.get(
            "colrev_status", "NA"
        ):
            return record_dict

        pdf_path = search_operation.review_manager.path / Path(record_dict["file"])
        tei = search_operation.review_manager.get_tei(
            pdf_path=pdf_path,
        )

        extracted_record = tei.get_metadata()

        for key, val in extracted_record.items():
            if val:
                record_dict[key] = str(val)

        with open(pdf_path, "rb") as file:
            parser = PDFParser(file)
            doc = PDFDocument(parser)

            if record_dict.get("title", "NA") in ["NA", ""]:
                if "Title" in doc.info[0]:
                    try:
                        record_dict["title"] = doc.info[0]["Title"].decode("utf-8")
                    except UnicodeDecodeError:
                        pass
            if record_dict.get("author", "NA") in ["NA", ""]:
                if "Author" in doc.info[0]:
                    try:
                        pdf_md_author = doc.info[0]["Author"].decode("utf-8")
                        if (
                            "Mirko Janc" not in pdf_md_author
                            and "wendy" != pdf_md_author
                            and "yolanda" != pdf_md_author
                        ):
                            record_dict["author"] = pdf_md_author
                    except UnicodeDecodeError:
                        pass

            if "abstract" in record_dict:
                del record_dict["abstract"]
            if "keywords" in record_dict:
                del record_dict["keywords"]

            # to allow users to update/reindex with newer version:
            record_dict["grobid-version"] = (
                "lfoppiano/grobid:" + tei.get_grobid_version()
            )

            return record_dict

    def __get_grobid_metadata(
        self, *, search_operation: colrev.ops.search.Search, pdf_path: Path
    ) -> dict:
        record_dict: typing.Dict[str, typing.Any] = {
            "file": str(pdf_path),
            "ENTRYTYPE": "misc",
        }
        try:
            record_dict = self.__get_record_from_pdf_grobid(
                search_operation=search_operation, record_dict=record_dict
            )

            with open(pdf_path, "rb") as file:
                parser = PDFParser(file)
                document = PDFDocument(parser)
                pages_in_file = resolve1(document.catalog["Pages"])["Count"]
                if pages_in_file < 6:
                    record = colrev.record.Record(data=record_dict)
                    record.set_text_from_pdf(
                        project_path=search_operation.review_manager.path
                    )
                    record_dict = record.get_data()
                    if "text_from_pdf" in record_dict:
                        text: str = record_dict["text_from_pdf"]
                        if "bookreview" in text.replace(" ", "").lower():
                            record_dict["ENTRYTYPE"] = "misc"
                            record_dict["note"] = "Book review"
                        if "erratum" in text.replace(" ", "").lower():
                            record_dict["ENTRYTYPE"] = "misc"
                            record_dict["note"] = "Erratum"
                        if "correction" in text.replace(" ", "").lower():
                            record_dict["ENTRYTYPE"] = "misc"
                            record_dict["note"] = "Correction"
                        if "contents" in text.replace(" ", "").lower():
                            record_dict["ENTRYTYPE"] = "misc"
                            record_dict["note"] = "Contents"
                        if "withdrawal" in text.replace(" ", "").lower():
                            record_dict["ENTRYTYPE"] = "misc"
                            record_dict["note"] = "Withdrawal"
                        del record_dict["text_from_pdf"]
                    # else:
                    #     print(f'text extraction error in {record_dict["ID"]}')
                    if "pages_in_file" in record_dict:
                        del record_dict["pages_in_file"]

                record_dict = {k: v for k, v in record_dict.items() if v is not None}
                record_dict = {k: v for k, v in record_dict.items() if v != "NA"}

                # add details based on path
                record_dict = self.__update_fields_based_on_pdf_dirs(
                    record_dict=record_dict, params=self.search_source.search_parameters
                )

        except colrev_exceptions.TEIException:
            pass

        return record_dict

    def __is_broken_filepath(
        self,
        pdf_path: Path,
    ) -> bool:
        if ";" in str(pdf_path):
            self.review_manager.logger.error(
                f'skipping PDF with ";" in filepath: \n{pdf_path}'
            )
            return True

        if (
            "_ocr.pdf" == str(pdf_path)[-8:]
            or "_wo_cp.pdf" == str(pdf_path)[-10:]
            or "_wo_lp.pdf" == str(pdf_path)[-10:]
            or "_backup.pdf" == str(pdf_path)[-11:]
        ):
            self.review_manager.logger.info(
                f"Skipping PDF with _ocr.pdf/_wo_cp.pdf: {pdf_path}"
            )
            return True

        return False

    def validate_source(
        self,
        search_operation: colrev.ops.search.Search,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Validate the SearchSource (parameters etc.)"""

        search_operation.review_manager.logger.debug(
            f"Validate SearchSource {source.filename}"
        )

        if "subdir_pattern" in source.search_parameters:
            if source.search_parameters["subdir_pattern"] != [
                "NA",
                "volume_number",
                "year",
                "volume",
            ]:
                raise colrev_exceptions.InvalidQueryException(
                    "subdir_pattern not in [NA, volume_number, year, volume]"
                )

        if "sub_dir_pattern" in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "sub_dir_pattern: deprecated. use subdir_pattern"
            )

        if "scope" not in source.search_parameters:
            raise colrev_exceptions.InvalidQueryException(
                "scope required in search_parameters"
            )
        if "path" not in source.search_parameters["scope"]:
            raise colrev_exceptions.InvalidQueryException(
                "path required in search_parameters/scope"
            )
        search_operation.review_manager.logger.debug(
            f"SearchSource {source.filename} validated"
        )

    def __add_md_string(self, *, record_dict: dict) -> dict:
        md_copy = record_dict.copy()
        try:
            fsize = str(
                (self.review_manager.path / Path(record_dict["file"])).stat().st_size
            )
        except FileNotFoundError:
            fsize = "NOT_FOUND"
        for key in ["ID", "grobid-version", "file"]:
            if key in md_copy:
                md_copy.pop(key)
        md_string = ",".join([f"{k}:{v}" for k, v in md_copy.items()])
        record_dict["md_string"] = str(fsize) + md_string
        return record_dict

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of a PDF directory (based on GROBID)"""

        # pylint: disable=too-many-locals
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        # Do not run in continuous-integration environment
        if search_operation.review_manager.in_ci_environment():
            return

        # Removing records/origins for which PDFs were removed makes sense for curated repositories
        # In regular repositories, it may be confusing (e.g., if PDFs are renamed)
        # In these cases, we may simply print a warning instead of modifying/removing records?
        if self.review_manager.settings.is_curated_masterdata_repo():
            self.__remove_records_if_pdf_no_longer_exists(
                search_operation=search_operation
            )

        pdfs_dir_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        records = search_operation.review_manager.dataset.load_records_dict()
        grobid_service = search_operation.review_manager.get_grobid_service()
        grobid_service.start()

        local_index = search_operation.review_manager.get_local_index()

        pdfs_to_index = [
            x.relative_to(search_operation.review_manager.path)
            for x in self.pdfs_path.glob("**/*.pdf")
        ]

        linked_pdf_paths = [Path(r["file"]) for r in records.values() if "file" in r]

        if search_operation.review_manager.force_mode:  # i.e., reindex all
            search_operation.review_manager.logger.info("Reindex all")

        batch_size = 20
        pdf_batches = [
            pdfs_to_index[i * batch_size : (i + 1) * batch_size]
            for i in range((len(pdfs_to_index) + batch_size - 1) // batch_size)
        ]
        nr_added, nr_changed = 0, 0
        for pdf_batch in pdf_batches:
            for record in pdfs_dir_feed.feed_records.values():
                record = self.__add_md_string(record_dict=record)

            for pdf_path in pdf_batch:
                if self.__is_broken_filepath(pdf_path=pdf_path):
                    continue

                if search_operation.review_manager.force_mode:
                    # i.e., reindex all
                    pass
                else:
                    # note: for curations, we want all pdfs indexed/merged separately,
                    # in other projects, it is generally sufficient if the pdf is linked
                    if not self.review_manager.settings.is_curated_masterdata_repo():
                        if pdf_path in linked_pdf_paths:
                            # Otherwise: skip linked PDFs
                            continue

                    if pdf_path in [
                        Path(r["file"])
                        for r in pdfs_dir_feed.feed_records.values()
                        if "file" in r
                    ]:
                        continue

                search_operation.review_manager.logger.info(
                    f" extract metadata from {pdf_path}"
                )
                try:
                    if (
                        not search_operation.review_manager.settings.is_curated_masterdata_repo()
                    ):
                        # retrieve_based_on_colrev_pdf_id
                        pdf_hash_service = (
                            search_operation.review_manager.get_pdf_hash_service()
                        )
                        colrev_pdf_id = pdf_hash_service.get_pdf_hash(
                            pdf_path=Path(pdf_path),
                            page_nr=1,
                            hash_size=32,
                        )
                        new_record = local_index.retrieve_based_on_colrev_pdf_id(
                            colrev_pdf_id="cpid1:" + colrev_pdf_id
                        )
                        new_record["file"] = str(pdf_path)
                        # Note : an alternative to replacing all data with the curated version
                        # is to just add the curation_ID
                        # (and retrieve the curated metadata separately/non-redundantly)
                    else:
                        new_record = self.__get_grobid_metadata(
                            search_operation=search_operation, pdf_path=pdf_path
                        )
                except (
                    colrev_exceptions.PDFHashError,
                    colrev_exceptions.RecordNotInIndexException,
                ):
                    # otherwise, get metadata from grobid (indexing)
                    new_record = self.__get_grobid_metadata(
                        search_operation=search_operation, pdf_path=pdf_path
                    )

                new_record = self.__add_md_string(record_dict=new_record)

                # Note: identical md_string as a heuristic for duplicates
                potential_duplicates = [
                    r
                    for r in pdfs_dir_feed.feed_records.values()
                    if r["md_string"] == new_record["md_string"]
                    and not r["file"] == new_record["file"]
                ]
                if potential_duplicates:
                    search_operation.review_manager.logger.warning(
                        f" {colors.RED}skip record (PDF potential duplicate): "
                        f"{new_record['file']} {colors.END} "
                        f"({','.join([r['file'] for r in potential_duplicates])})"
                    )
                    continue

                try:
                    pdfs_dir_feed.set_id(record_dict=new_record)
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

                prev_record_dict_version = {}
                if new_record["ID"] in pdfs_dir_feed.feed_records:
                    prev_record_dict_version = pdfs_dir_feed.feed_records[
                        new_record["ID"]
                    ]

                added = pdfs_dir_feed.add_record(
                    record=colrev.record.Record(data=new_record),
                )
                if added:
                    nr_added += 1

                    if "doi" not in new_record:
                        record = colrev.record.Record(data=new_record)
                        record.set_text_from_pdf(
                            project_path=search_operation.review_manager.path
                        )
                        res = re.findall(self.__doi_regex, record.data["text_from_pdf"])
                        if res:
                            record.data["doi"] = res[0].upper()
                        del record.data["text_from_pdf"]

                elif search_operation.review_manager.force_mode:
                    # Note : only re-index/update
                    changed = search_operation.update_existing_record(
                        records=records,
                        record_dict=new_record,
                        prev_record_dict_version=prev_record_dict_version,
                        source=self.search_source,
                        update_time_variant_fields=rerun,
                    )
                    if changed:
                        nr_changed += 1

            for record in pdfs_dir_feed.feed_records.values():
                record.pop("md_string")

            pdfs_dir_feed.save_feed_file()

        if nr_added > 0:
            search_operation.review_manager.logger.info(
                f"{colors.GREEN}Retrieved {nr_added} records{colors.END}"
            )
        else:
            search_operation.review_manager.logger.info(
                f"{colors.GREEN}No additional records retrieved{colors.END}"
            )

        if search_operation.review_manager.force_mode:
            if nr_changed > 0:
                search_operation.review_manager.logger.info(
                    f"{colors.GREEN}Updated {nr_changed} records{colors.END}"
                )
            else:
                if records:
                    search_operation.review_manager.logger.info(
                        f"{colors.GREEN}Records (data/records.bib) up-to-date{colors.END}"
                    )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for PDF directories (GROBID)"""

        result = {"confidence": 0.0}

        if filename.suffix == ".pdf" and not bws.BackwardSearchSource.heuristic(
            filename=filename, data=data
        ):
            result["confidence"] = 1.0
            return result

        return result

    @classmethod
    def add_endpoint(
        cls, search_operation: colrev.ops.search.Search, query: str
    ) -> typing.Optional[colrev.settings.SearchSource]:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        if "pdfs" == query:
            filename = search_operation.get_unique_filename(file_path_string="pdfs")
            # pylint: disable=no-value-for-parameter
            add_source = colrev.settings.SearchSource(
                endpoint="colrev.pdfs_dir",
                filename=filename,
                search_type=colrev.settings.SearchType.PDFS,
                search_parameters={"scope": {"path": "data/pdfs"}},
                load_conversion_package_endpoint={"endpoint": "colrev.bibtex"},
                comment="",
            )
            return add_source

        return None

    def load_fixes(
        self,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
        records: typing.Dict,
    ) -> dict:
        """Load fixes for PDF directories (GROBID)"""

        for record in records.values():
            if "grobid-version" in record:
                del record["grobid-version"]

            if "doi" in record:
                try:
                    retrieved_record = self.crossref_connector.query_doi(
                        doi=record["doi"]
                    )

                    for key in [
                        "journal",
                        "booktitle",
                        "volume",
                        "number",
                        "year",
                        "pages",
                    ]:
                        if key in retrieved_record.data:
                            record[key] = retrieved_record.data[key]
                except colrev_exceptions.RecordNotFoundInPrepSourceException:
                    pass

        return records

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for PDF directories (GROBID)"""

        # Typical error in old papers: title fields are equal to journal/booktitle fields
        if record.data.get("title", "no_title").lower() == record.data.get(
            "journal", "no_journal"
        ):
            record.remove_field(key="title", source="pdfs_dir_prepare")
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )
        if record.data.get("title", "no_title").lower() == record.data.get(
            "booktitle", "no_booktitle"
        ):
            record.remove_field(key="title", source="pdfs_dir_prepare")
            record.set_status(
                target_state=colrev.record.RecordState.md_needs_manual_preparation
            )

        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Load conversion of tables (xlsx, csv)"""
from __future__ import annotations

from dataclasses import asdict
from dataclasses import dataclass

import pandas as pd
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.load

# pylint: disable=too-few-public-methods
# pylint: disable=unused-argument
# pylint: disable=duplicate-code


class TableLoadUtility:
    """Utility for tables loading"""

    @classmethod
    def preprocess_records(cls, *, records: list) -> dict:
        """Preprocess records imported from a table"""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        next_id = 1
        for record_dict in records:
            if "type" in record_dict:
                record_dict["ENTRYTYPE"] = record_dict["type"]
                del record_dict["type"]
                if "inproceedings" == record_dict["ENTRYTYPE"]:
                    if "journal" in record_dict and "booktitle" not in record_dict:
                        record_dict["booktitle"] = record_dict["journal"]
                        del record_dict["journal"]
                if "article" == record_dict["ENTRYTYPE"]:
                    if "booktitle" in record_dict and "journal" not in record_dict:
                        record_dict["journal"] = record_dict["booktitle"]
                        del record_dict["booktitle"]

            if "ENTRYTYPE" not in record_dict:
                if "" != record_dict.get("journal", ""):
                    record_dict["ENTRYTYPE"] = "article"
                if "" != record_dict.get("booktitle", ""):
                    record_dict["ENTRYTYPE"] = "inproceedings"
                else:
                    record_dict["ENTRYTYPE"] = "misc"

            if "ID" not in record_dict:
                if "citation_key" in record_dict:
                    record_dict["ID"] = record_dict["citation_key"]
                else:
                    record_dict["ID"] = next_id
                    next_id += 1

            if "issue" in record_dict and "number" not in record_dict:
                record_dict["number"] = record_dict["issue"]
                if "no issue" == record_dict["number"]:
                    del record_dict["number"]
                del record_dict["issue"]

            for key, value in record_dict.items():
                record_dict[key] = str(value)

            if "authors" in record_dict and "author" not in record_dict:
                record_dict["author"] = record_dict["authors"]
                del record_dict["authors"]
            if "publication_year" in record_dict and "year" not in record_dict:
                record_dict["year"] = record_dict["publication_year"]
                del record_dict["publication_year"]
            # Note: this is a simple heuristic:
            if (
                "journal/book" in record_dict
                and "journal" not in record_dict
                and "doi" in record_dict
            ):
                record_dict["journal"] = record_dict["journal/book"]
                del record_dict["journal/book"]

            if "no Times-Cited" == record_dict.get("cited_by", ""):
                del record_dict["cited_by"]

        if all("ID" in r for r in records):
            records_dict = {r["ID"]: r for r in records}
        else:
            records_dict = {}
            for i, record in enumerate(records):
                records_dict[str(i)] = record

        for r_dict in records_dict.values():
            if "no year" == r_dict.get("year", "NA"):
                del r_dict["year"]
            if "no journal" == r_dict.get("journal", "NA"):
                del r_dict["journal"]
            if "no volume" == r_dict.get("volume", "NA"):
                del r_dict["volume"]
            if "no pages" == r_dict.get("pages", "NA"):
                del r_dict["pages"]
            if "no issue" == r_dict.get("issue", "NA"):
                del r_dict["issue"]
            if "no number" == r_dict.get("number", "NA"):
                del r_dict["number"]
            if "no doi" == r_dict.get("doi", "NA"):
                del r_dict["doi"]
            if "no type" == r_dict.get("type", "NA"):
                del r_dict["type"]
            if "author_count" in r_dict:
                del r_dict["author_count"]
            if "no Number-of-Cited-References" == r_dict.get(
                "number_of_cited_references", "NA"
            ):
                del r_dict["number_of_cited_references"]
            if "no file" in r_dict.get("file_name", "NA"):
                del r_dict["file_name"]
            if "times_cited" == r_dict.get("times_cited", "NA"):
                del r_dict["times_cited"]

        return records_dict


@zope.interface.implementer(
    colrev.env.package_manager.LoadConversionPackageEndpointInterface
)
@dataclass
class CSVLoader(JsonSchemaMixin):

    """Loads csv files (based on pandas)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    supported_extensions = ["csv"]

    def __init__(
        self,
        *,
        load_operation: colrev.ops.load.Load,
        settings: dict,
    ):
        self.settings = self.settings_class.load_settings(data=settings)

    def load(
        self, load_operation: colrev.ops.load.Load, source: colrev.settings.SearchSource
    ) -> dict:
        """Load records from the source"""

        try:
            data = pd.read_csv(source.filename)
        except pd.errors.ParserError as exc:
            raise colrev_exceptions.ImportException(
                f"Error: Not a csv file? {source.filename.name}"
            ) from exc

        data.columns = data.columns.str.replace(" ", "_")
        data.columns = data.columns.str.replace("-", "_")
        data.columns = data.columns.str.lower()
        records_value_list = data.to_dict("records")

        records_dict = TableLoadUtility.preprocess_records(records=records_value_list)

        if not all("ID" in r for r in records_dict.values()):
            for i, record in enumerate(records_dict.values()):
                record["ID"] = str(i).rjust(6, "0")

        records = {r["ID"]: r for r in records_dict.values()}

        endpoint_dict = load_operation.package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            selected_packages=[asdict(source)],
            operation=load_operation,
            ignore_not_available=False,
        )
        endpoint = endpoint_dict[source.endpoint]

        records = endpoint.load_fixes(  # type: ignore
            load_operation, source=source, records=records
        )

        return records


@zope.interface.implementer(
    colrev.env.package_manager.LoadConversionPackageEndpointInterface
)
class ExcelLoader:
    """Loads Excel (xls, xlsx) files (based on pandas)"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    supported_extensions = ["xls", "xlsx"]

    def __init__(
        self,
        *,
        load_operation: colrev.ops.load.Load,
        settings: dict,
    ):
        self.settings = self.settings_class.load_settings(data=settings)

    def load(
        self, load_operation: colrev.ops.load.Load, source: colrev.settings.SearchSource
    ) -> dict:
        """Load records from the source"""

        try:
            data = pd.read_excel(
                source.filename, dtype=str
            )  # dtype=str to avoid type casting
        except pd.errors.ParserError:
            load_operation.review_manager.logger.error(
                f"Error: Not an xlsx file: {source.filename.name}"
            )
            return {}

        data.columns = data.columns.str.replace(" ", "_")
        data.columns = data.columns.str.replace("-", "_")
        data.columns = data.columns.str.lower()
        record_value_list = data.to_dict("records")
        records_dicts = TableLoadUtility.preprocess_records(records=record_value_list)
        records = {r["ID"]: r for r in records_dicts}

        endpoint_dict = load_operation.package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            selected_packages=[asdict(source)],
            operation=load_operation,
            ignore_not_available=False,
        )
        endpoint = endpoint_dict[source.endpoint]

        records = endpoint.load_fixes(  # type: ignore
            load_operation, source=source, records=records
        )
        return records


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Export of bib/pdfs as a prep-man operation"""
from __future__ import annotations

import os
import subprocess
import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin
from PyPDF2 import PdfFileReader
from PyPDF2 import PdfFileWriter

import colrev.env.package_manager
import colrev.env.utils
import colrev.record

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.prep_man

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepManPackageEndpointInterface)
@dataclass
class ExportManPrep(JsonSchemaMixin):
    """Manual preparation based on exported and imported metadata (and PDFs if any)"""

    ci_supported: bool = False

    RELATIVE_PREP_MAN_PATH = Path("records_prep_man.bib")

    @dataclass
    class ExportManPrepSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for ExportManPrep"""

        endpoint: str
        pdf_handling_mode: str = "symlink"

        _details = {
            "pdf_handling_mode": {
                "tooltip": "Indicates how linked PDFs are handled (symlink/copy_first_page)"
            },
        }

    settings_class = ExportManPrepSettings

    def __init__(
        self,
        *,
        prep_man_operation: colrev.ops.prep_man.PrepMan,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        if "pdf_handling_mode" not in settings:
            settings["pdf_handling_mode"] = "symlink"
        assert settings["pdf_handling_mode"] in ["symlink", "copy_first_page"]

        self.settings = self.settings_class.load_settings(data=settings)

        self.review_manager = prep_man_operation.review_manager

        self.prep_man_bib_path = (
            self.review_manager.prep_dir / self.RELATIVE_PREP_MAN_PATH
        )

        self.review_manager.prep_dir.mkdir(exist_ok=True, parents=True)

    def __copy_files_for_man_prep(self, *, records: dict) -> None:
        prep_man_path_pdfs = self.review_manager.prep_dir / Path("pdfs")
        if prep_man_path_pdfs.is_dir():
            input(f"Remove {prep_man_path_pdfs} and press Enter.")
        prep_man_path_pdfs.mkdir(exist_ok=True, parents=True)

        for record in records.values():
            if "file" in record:
                target_path = self.review_manager.prep_dir / Path(record["file"])
                target_path.parents[0].mkdir(exist_ok=True, parents=True)

                if "symlink" == self.settings.pdf_handling_mode:
                    target_path.symlink_to(Path(record["file"]).resolve())

                if "copy_first_page" == self.settings.pdf_handling_mode:
                    pdf_reader = PdfFileReader(str(record["file"]), strict=False)
                    if len(pdf_reader.pages) >= 1:
                        writer = PdfFileWriter()
                        writer.addPage(pdf_reader.getPage(0))
                        with open(target_path, "wb") as outfile:
                            writer.write(outfile)

    def __export_prep_man(
        self,
        *,
        prep_man_operation: colrev.ops.prep_man.PrepMan,
        records: typing.Dict[str, typing.Dict],
    ) -> None:
        prep_man_operation.review_manager.logger.info(
            f"Export records for man-prep to {self.prep_man_bib_path}"
        )

        man_prep_recs = {
            k: v
            for k, v in records.items()
            if colrev.record.RecordState.md_needs_manual_preparation
            == v["colrev_status"]
        }
        prep_man_operation.review_manager.dataset.save_records_dict_to_file(
            records=man_prep_recs, save_path=self.prep_man_bib_path
        )
        if any("file" in r for r in man_prep_recs.values()):
            self.__copy_files_for_man_prep(records=man_prep_recs)
        if "pytest" not in os.getcwd():
            # os.system('%s %s' % (os.getenv('EDITOR'), self.prep_man_bib_path))
            subprocess.call(["xdg-open", str(self.prep_man_bib_path)])

    def __import_prep_man(
        self, *, prep_man_operation: colrev.ops.prep_man.PrepMan
    ) -> None:
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals

        prep_man_operation.review_manager.logger.info(
            "Load import changes from "
            f"{self.prep_man_bib_path.relative_to(prep_man_operation.review_manager.path)}"
        )

        with open(self.prep_man_bib_path, encoding="utf8") as target_bib:
            man_prep_recs = prep_man_operation.review_manager.dataset.load_records_dict(
                load_str=target_bib.read()
            )

        imported_records = []
        records = prep_man_operation.review_manager.dataset.load_records_dict()
        for record_id, record_dict in man_prep_recs.items():
            if record_id not in records:
                print(f"ID no longer in records: {record_id}")
                continue
            record = colrev.record.PrepRecord(data=record_dict)
            record.update_masterdata_provenance()
            record.set_status(target_state=colrev.record.RecordState.md_prepared)
            if colrev.record.RecordState.md_prepared == record.data["colrev_status"]:
                imported_records.append(record.data["ID"])
            for k in list(record.data.keys()):
                if k in [
                    "colrev_status",
                    "colrev_masterdata_provenance",
                    "colrev_data_provenance",
                    "colrev_id",
                ]:
                    continue
                if k in records[record_id]:
                    if record.data[k] != records[record_id][k]:
                        if k in record.data.get("colrev_masterdata_provenance", {}):
                            record.add_masterdata_provenance(key=k, source="man_prep")
                        else:
                            record.add_data_provenance(key=k, source="man_prep")
                else:
                    if k in records[record_id]:
                        del records[record_id][k]
                    if k in record.data.get("colrev_masterdata_provenance", {}):
                        record.add_masterdata_provenance(
                            key=k, source="man_prep", note="not_missing"
                        )
                    else:
                        record.add_data_provenance(
                            key=k, source="man_prep", note="not_missing"
                        )
            colrev_data_provenance_keys_to_drop = []
            for key, items in record.data.get("colrev_data_provenance", {}).items():
                if key not in record.data and "not_missing" not in items["note"]:
                    colrev_data_provenance_keys_to_drop.append(key)
            for (
                colrev_data_provenance_key_to_drop
            ) in colrev_data_provenance_keys_to_drop:
                del record.data["colrev_data_provenance"][
                    colrev_data_provenance_key_to_drop
                ]

            colrev_masterdata_provenance_keys_to_drop = []
            for key, items in record.data.get(
                "colrev_masterdata_provenance", {}
            ).items():
                if key not in record.data and "not_missing" not in items["note"]:
                    colrev_masterdata_provenance_keys_to_drop.append(key)
            for (
                colrev_masterdata_provenance_key_to_drop
            ) in colrev_masterdata_provenance_keys_to_drop:
                del record.data["colrev_masterdata_provenance"][
                    colrev_masterdata_provenance_key_to_drop
                ]

            records[record_id] = record.get_data()

        prep_man_operation.review_manager.dataset.save_records_dict(records=records)
        prep_man_operation.review_manager.dataset.add_record_changes()
        prep_man_operation.review_manager.create_commit(msg="Prep-man (ExportManPrep)")

        prep_man_operation.review_manager.dataset.set_ids(selected_ids=imported_records)
        prep_man_operation.review_manager.create_commit(msg="Set IDs")

    def prepare_manual(
        self, prep_man_operation: colrev.ops.prep_man.PrepMan, records: dict
    ) -> dict:
        """Prepare records manually by extracting the subset of records to a separate BiBTex file"""

        if not self.prep_man_bib_path.is_file():
            self.__export_prep_man(
                prep_man_operation=prep_man_operation, records=records
            )
        else:
            selected_path = self.prep_man_bib_path.relative_to(
                prep_man_operation.review_manager.path
            )
            if "y" == input(f"Import changes from {selected_path} [y,n]?"):
                self.__import_prep_man(prep_man_operation=prep_man_operation)

        return records


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Manages environment registry, services, and stauts"""
from __future__ import annotations

import json
import typing
from pathlib import Path
from subprocess import CalledProcessError
from subprocess import check_output
from typing import Optional

import docker
import git
import pandas as pd
import yaml
from docker.errors import DockerException
from yaml import safe_load

import colrev.env.local_index
import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record
import colrev.ui_cli.cli_colors as colors


class EnvironmentManager:
    """The EnvironmentManager manages environment resources and services"""

    colrev_path = Path.home().joinpath("colrev")
    cache_path = colrev_path / Path("prep_requests_cache")
    REGISTRY_RELATIVE = Path("registry.yaml")
    registry = colrev_path.joinpath(REGISTRY_RELATIVE)

    def __init__(self) -> None:
        self.environment_registry = self.load_environment_registry()
        self.__registered_ports: typing.List[str] = []
        self.__registered_services: typing.List[str] = []

    def register_ports(self, *, ports: typing.List[str]) -> None:
        """Register a localhost port to avoid conflicts"""
        for port_to_register in ports:
            if port_to_register in self.__registered_ports:
                raise colrev_exceptions.PortAlreadyRegisteredException(
                    f"Port {port_to_register} already registered"
                )
            self.__registered_ports.append(port_to_register)

    def register_docker_service(self, *, imagename: str) -> None:
        """Register a docker service"""
        self.__registered_services.append(imagename)

    def stop_docker_services(self) -> None:
        """Stop registered docker services"""

        try:
            client = docker.from_env()
            for container in client.containers.list():
                if any(x in str(container.image) for x in self.__registered_services):
                    container.stop()
                    print(f"Stopped container {container.name} ({container.image})")
        except DockerException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Docker service not available ({exc}). Please install/start Docker."
            ) from exc

    def load_environment_registry(self) -> list:
        """Load the local registry"""
        environment_registry_path = self.registry
        environment_registry = []
        if environment_registry_path.is_file():
            with open(environment_registry_path, encoding="utf8") as file:
                environment_registry_df = pd.json_normalize(safe_load(file))
                environment_registry = environment_registry_df.to_dict("records")

        return environment_registry

    def save_environment_registry(self, *, updated_registry: list) -> None:
        """Save the local registry"""
        updated_registry_df = pd.DataFrame(updated_registry)
        ordered_cols = [
            "repo_name",
            "repo_source_path",
        ]
        for entry in [x for x in updated_registry_df.columns if x not in ordered_cols]:
            ordered_cols.append(entry)
        updated_registry_df = updated_registry_df.reindex(columns=ordered_cols)

        self.registry.parents[0].mkdir(parents=True, exist_ok=True)
        with open(self.registry, "w", encoding="utf8") as file:
            yaml.dump(
                json.loads(
                    updated_registry_df.to_json(orient="records", default_handler=str)
                ),
                file,
                default_flow_style=False,
                sort_keys=False,
            )

    def register_repo(self, *, path_to_register: Path) -> None:
        """Register a repository"""
        self.environment_registry = self.load_environment_registry()
        registered_paths = [x["repo_source_path"] for x in self.environment_registry]

        if registered_paths != []:
            if str(path_to_register) in registered_paths:
                # print(f"Warning: Path already registered: {path_to_register}")
                return
        else:
            print(f"Creating {self.registry}")

        new_record = {
            "repo_name": path_to_register.stem,
            "repo_source_path": path_to_register,
        }
        git_repo = git.Repo(path_to_register)
        for remote in git_repo.remotes:
            if remote.url:
                new_record["repo_source_url"] = remote.url
        self.environment_registry.append(new_record)
        self.save_environment_registry(updated_registry=self.environment_registry)
        print(f"Registered path ({path_to_register})")

    def get_name_mail_from_git(self) -> typing.Tuple[str, str]:  # pragma: no cover
        """Get the committer name and email from git (globals)"""
        global_conf_details = ("NA", "NA")
        try:
            username = check_output(["git", "config", "user.name"])
            email = check_output(["git", "config", "user.email"])
            global_conf_details = (
                username.decode("utf-8").replace("\n", ""),
                email.decode("utf-8").replace("\n", ""),
            )
        except CalledProcessError as exc:
            raise colrev_exceptions.CoLRevException(
                "Global git variables (user name and email) not available."
            ) from exc
        if ("NA", "NA") == global_conf_details:
            raise colrev_exceptions.CoLRevException(
                "Global git variables (user name and email) not available."
            )
        return global_conf_details

    @classmethod
    def build_docker_image(
        cls, *, imagename: str, image_path: Optional[Path] = None
    ) -> None:
        """Build a docker image"""

        try:
            client = docker.from_env()
            repo_tags = [t for image in client.images.list() for t in image.tags]

            if imagename not in repo_tags:
                if image_path:
                    assert colrev.review_manager.__file__
                    colrev_path = Path("")
                    if colrev.review_manager.__file__:
                        colrev_path = Path(colrev.review_manager.__file__).parents[0]
                    print(f"Building {imagename} Docker image ...")
                    context_path = colrev_path / image_path
                    client.images.build(
                        path=str(context_path), tag=f"{imagename}:latest"
                    )

                else:
                    print(f"Pulling {imagename} Docker image...")
                    client.images.pull(imagename)
        except DockerException as exc:
            raise colrev_exceptions.ServiceNotAvailableException(
                dep="docker",
                detailed_trace=f"Docker service not available ({exc}). "
                + "Please install/start Docker.",
            ) from exc

    def check_git_installed(self) -> None:
        """Check whether git is installed"""

        try:
            git_instance = git.Git()
            _ = git_instance.version()
        except Exception as exc:  # pylint: disable=broad-except
            print(exc)
            # raise colrev_exceptions.MissingDependencyError("git") from exc

    def check_docker_installed(self) -> None:
        """Check whether Docker is installed"""

        try:
            client = docker.from_env()
            _ = client.version()
        except docker.errors.DockerException as exc:
            if "PermissionError" in exc.args[0]:
                raise colrev_exceptions.DependencyConfigurationError(
                    "Docker: Permission error. Run "
                    f"{colors.ORANGE}sudo gpasswd -a $USER docker && "
                    f"newgrp docker{colors.END}"
                )
            raise colrev_exceptions.MissingDependencyError("Docker")

    def _get_status(
        self, *, review_manager: colrev.review_manager.ReviewManager
    ) -> dict:
        status_dict = {}
        with open(review_manager.status, encoding="utf8") as stream:
            try:
                status_dict = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
        return status_dict

    def get_environment_details(self) -> dict:
        """Get the environment details"""

        # def get_last_modified() -> str:

        #     list_of_files = local_index.opensearch_index.glob(
        #         "**/*"
        #     )  # * means all if need specific format then *.csv
        #     latest_file = max(list_of_files, key=os.path.getmtime)
        #     last_mod = datetime.fromtimestamp(latest_file.lstat().st_mtime)
        #     return last_mod.strftime("%Y-%m-%d %H:%M")

        local_index = colrev.env.local_index.LocalIndex(
            index_tei=True, verbose_mode=True
        )

        environment_details = {}
        size = 0
        last_modified = "NOT_INITIATED"
        status = "TODO"

        size = 0
        # try:
        #     size = local_index.open_search.cat.count(
        #         index=local_index.RECORD_INDEX, params={"format": "json"}
        #     )[0]["count"]
        #     last_modified = get_last_modified()
        #     status = "up"
        # except (NotFoundError, IndexError):
        #     status = "down"

        environment_details["index"] = {
            "size": size,
            "last_modified": last_modified,
            "path": str(local_index.local_environment_path),
            "status": status,
        }

        environment_stats = self.get_environment_stats()

        environment_details["local_repos"] = {
            "repos": environment_stats["repos"],
            "broken_links": environment_stats["broken_links"],
        }
        return environment_details

    def get_environment_stats(self) -> dict:
        """Get the environment stats"""
        local_repos = self.load_environment_registry()
        repos = []
        broken_links = []
        for repo in local_repos:
            try:
                cp_review_manager = colrev.review_manager.ReviewManager(
                    path_str=repo["repo_source_path"]
                )
                check_operation = colrev.operation.CheckOperation(
                    review_manager=cp_review_manager
                )
                repo_stat = self._get_status(review_manager=cp_review_manager)
                repo["size"] = repo_stat["overall"]["md_processed"]
                if repo_stat["atomic_steps"] != 0:
                    repo["progress"] = round(
                        repo_stat["completed_atomic_steps"] / repo_stat["atomic_steps"],
                        2,
                    )
                else:
                    repo["progress"] = -1

                repo["remote"] = False
                git_repo = check_operation.review_manager.dataset.get_repo()
                for remote in git_repo.remotes:
                    if remote.url:
                        repo["remote"] = True
                repo[
                    "behind_remote"
                ] = check_operation.review_manager.dataset.behind_remote()

                repos.append(repo)
            except colrev_exceptions.CoLRevException:
                broken_links.append(repo)
        return {"repos": repos, "broken_links": broken_links}

    def get_curated_outlets(self) -> list:
        """Get the curated outlets"""
        curated_outlets: typing.List[str] = []
        for repo_source_path in [
            x["repo_source_path"]
            for x in self.load_environment_registry()
            if "colrev/curated_metadata/" in x["repo_source_path"]
        ]:
            try:
                # TODO : should come from the curation_data endpoint!
                with open(f"{repo_source_path}/readme.md", encoding="utf-8") as file:
                    first_line = file.readline()
                curated_outlets.append(first_line.lstrip("# ").replace("\n", ""))

                with open(
                    f"{repo_source_path}/data/records.bib", encoding="utf-8"
                ) as file:
                    outlets = []
                    for line in file.readlines():
                        # Note : the second part ("journal:"/"booktitle:")
                        # ensures that data provenance fields are skipped
                        if (
                            "journal" == line.lstrip()[:7]
                            and "journal:" != line.lstrip()[:8]
                        ):
                            journal = line[line.find("{") + 1 : line.rfind("}")]
                            if "UNKNOWN" != journal:
                                outlets.append(journal)
                        if (
                            "booktitle" == line.lstrip()[:9]
                            and "booktitle:" != line.lstrip()[:10]
                        ):
                            booktitle = line[line.find("{") + 1 : line.rfind("}")]
                            if "UNKNOWN" != booktitle:
                                outlets.append(booktitle)

                    if len(set(outlets)) > 1:
                        raise colrev_exceptions.CuratedOutletNotUnique(
                            "Error: Duplicate outlets in curated_metadata of "
                            f"{repo_source_path} : {','.join(list(set(outlets)))}"
                        )
            except FileNotFoundError as exc:
                print(exc)

        return curated_outlets


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""CoLRev load operation: Load records from search sources into references.bib."""
from __future__ import annotations

import html
import itertools
import re
import string
import typing
from pathlib import Path

import colrev.exceptions as colrev_exceptions
import colrev.operation
import colrev.record
import colrev.settings
import colrev.ui_cli.cli_colors as colors

# pylint: disable=too-many-lines


class Load(colrev.operation.Operation):
    """Load the records"""

    # Note : PDFs should be stored in the pdfs directory
    # They should be included through colrev search

    supported_extensions: typing.List[str]

    __LATEX_SPECIAL_CHAR_MAPPING = {
        '\\"u': "ü",
        "\\&": "&",
        '\\"o': "ö",
        '\\"a': "ä",
        '\\"A': "Ä",
        '\\"O': "Ö",
        '\\"U': "Ü",
        "\\textendash": "–",
        "\\textemdash": "—",
        "\\~a": "ã",
        "\\'o": "ó",
    }

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        notify_state_transition_operation: bool = True,
        hide_load_explanation: bool = False,
    ) -> None:
        super().__init__(
            review_manager=review_manager,
            operations_type=colrev.operation.OperationsType.load,
            notify_state_transition_operation=notify_state_transition_operation,
        )

        self.package_manager = self.review_manager.get_package_manager()

        if not hide_load_explanation:
            self.review_manager.logger.info("Load")
            self.review_manager.logger.info(
                "Load converts search results and adds them to the shared data/records.bib."
            )
            self.review_manager.logger.info(
                "Original records (search results) are stored in the directory data/search"
            )
            self.review_manager.logger.info(
                "See https://colrev.readthedocs.io/en/latest/manual/metadata_retrieval/load.html"
            )

    def __get_new_search_files(self) -> list[Path]:
        """Retrieve new search files (not yet registered in settings)"""

        if not self.review_manager.search_dir.is_dir():
            return []

        files = [
            f.relative_to(self.review_manager.path)
            for f in self.review_manager.search_dir.glob("**/*")
        ]

        # Only files that are not yet registered
        # (also exclude bib files corresponding to a registered file)
        files = [
            f
            for f in files
            if str(f.with_suffix(".bib"))
            not in [
                str(s.filename.with_suffix(".bib"))
                for s in self.review_manager.settings.sources
            ]
        ]

        return sorted(list(set(files)))

    def __get_currently_imported_origin_list(self) -> list:
        records_headers = self.review_manager.dataset.load_records_dict(
            header_only=True
        )
        record_header_list = list(records_headers.values())
        imported_origins = [
            item for x in record_header_list for item in x["colrev_origin"]
        ]
        return imported_origins

    def __apply_source_heuristics(
        self,
        *,
        filepath: Path,
        search_sources: dict,
        load_conversion: dict,
    ) -> list[typing.Dict]:
        """Apply heuristics to identify source"""

        # pylint: disable=too-many-statements

        def get_load_conversion_package_endpoint(*, filepath: Path) -> dict:
            filetype = filepath.suffix.replace(".", "")

            for (
                package_identifier,
                selected_package,
            ) in load_conversion.items():
                if filetype in selected_package.supported_extensions:  # type: ignore
                    return {"endpoint": package_identifier}

            raise colrev_exceptions.UnsupportedImportFormatError(filepath)

        data = ""
        try:
            data = filepath.read_text()
        except UnicodeDecodeError:
            pass

        results_list = []
        for (
            endpoint,
            endpoint_class,
        ) in search_sources.items():
            # pylint: disable=no-member
            has_heuristic = getattr(endpoint_class, "heuristic", None)
            if not has_heuristic:
                self.review_manager.logger.debug(f"- {endpoint}: no heuristic")
                continue
            res = endpoint_class.heuristic(filepath, data)  # type: ignore
            self.review_manager.logger.debug(f"- {endpoint}: {res['confidence']}")
            try:
                if res["confidence"] > 0:
                    result_item = {}

                    res["endpoint"] = endpoint

                    search_type = endpoint_class.search_type
                    # Note : as the identifier, we use the filename
                    # (if search results are added by file/not via the API)

                    # Correct the file extension if necessary
                    if re.findall(
                        r"^%0", data, re.MULTILINE
                    ) and filepath.suffix not in [".enl"]:
                        new_filename = filepath.with_suffix(".enl")
                        self.review_manager.logger.info(
                            f"{colors.GREEN}Rename to {new_filename} "
                            f"(because the format is .enl){colors.END}"
                        )
                        filepath.rename(new_filename)
                        self.review_manager.dataset.add_changes(
                            path=filepath, remove=True
                        )
                        filepath = new_filename
                        res["filename"] = filepath
                        self.review_manager.dataset.add_changes(path=new_filename)
                        self.review_manager.create_commit(msg=f"Rename {filepath}")

                    if re.findall(
                        r"^TI ", data, re.MULTILINE
                    ) and filepath.suffix not in [".ris"]:
                        new_filename = filepath.with_suffix(".ris")
                        self.review_manager.logger.info(
                            f"{colors.GREEN}Rename to {new_filename} "
                            f"(because the format is .ris){colors.END}"
                        )
                        filepath.rename(new_filename)
                        self.review_manager.dataset.add_changes(
                            path=filepath, remove=True
                        )
                        filepath = new_filename
                        res["filename"] = filepath
                        self.review_manager.dataset.add_changes(path=new_filename)
                        self.review_manager.create_commit(msg=f"Rename {filepath}")

                    if "load_conversion_package_endpoint" not in res:
                        res[
                            "load_conversion_package_endpoint"
                        ] = get_load_conversion_package_endpoint(filepath=filepath)

                    source_candidate = colrev.settings.SearchSource(
                        endpoint=endpoint,
                        filename=filepath,
                        search_type=search_type,
                        search_parameters={},
                        load_conversion_package_endpoint=res[
                            "load_conversion_package_endpoint"
                        ],
                        comment="",
                    )

                    result_item["source_candidate"] = source_candidate
                    result_item["confidence"] = res["confidence"]

                    results_list.append(result_item)
            except colrev_exceptions.UnsupportedImportFormatError:
                continue

        # Reduce the results_list when there are results with very high confidence
        if [r for r in results_list if r["confidence"] > 0.95]:
            results_list = [r for r in results_list if r["confidence"] > 0.8]

        if (
            0 == len(results_list)
            or len([r for r in results_list if r["confidence"] > 0.5]) == 0
        ):
            source_candidate = colrev.settings.SearchSource(
                endpoint="colrev.unknown_source",
                filename=Path(filepath),
                search_type=colrev.settings.SearchType("DB"),
                search_parameters={},
                load_conversion_package_endpoint=get_load_conversion_package_endpoint(
                    filepath=filepath
                ),
                comment="",
            )
            results_list.append(
                {
                    "source_candidate": source_candidate,
                    "confidence": 0.1,  # type: ignore
                }
            )

        return results_list

    def get_new_sources(
        self, *, skip_query: bool = False
    ) -> typing.List[colrev.settings.SearchSource]:
        """Get new SearchSources"""

        # pylint: disable=redefined-outer-name
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements
        # pylint: disable=too-many-locals

        new_search_files = self.__get_new_search_files()
        if not new_search_files:
            self.review_manager.logger.info("No new search files...")
            return []

        self.review_manager.logger.debug("Load available search_source endpoints...")

        search_source_identifiers = self.package_manager.discover_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            installed_only=True,
        )

        search_sources = self.package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            selected_packages=[{"endpoint": p} for p in search_source_identifiers],
            operation=self,
            instantiate_objects=False,
        )

        self.review_manager.logger.debug("Load available load_conversion endpoints...")
        load_conversion_package_identifiers = self.package_manager.discover_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.load_conversion,
            installed_only=True,
        )

        load_conversion_packages = self.package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.load_conversion,
            selected_packages=[
                {"endpoint": p} for p in load_conversion_package_identifiers
            ],
            operation=self,
        )

        self.supported_extensions = [
            item
            for sublist in [
                e.supported_extensions  # type: ignore
                for _, e in load_conversion_packages.items()
            ]
            for item in sublist
        ]

        new_sources = []
        for sfp in new_search_files:
            sfp_name = sfp
            if sfp_name in [
                str(source.filename) for source in self.review_manager.settings.sources
            ]:
                continue

            if not self.review_manager.high_level_operation:
                print()
            self.review_manager.logger.info(f"Discover new source: {sfp_name}")

            # Assuming that all other search types are added by query
            # search_type_input = "NA"
            # while search_type_input not in SearchType.get_options():
            #     print(f"Search type options: {SearchType.get_options()}")
            #     cmd = "Enter search type".ljust(25, " ") + ": "
            #     search_type_input = input(cmd)

            heuristic_result_list = self.__apply_source_heuristics(
                filepath=sfp,
                search_sources=search_sources,
                load_conversion=load_conversion_packages,
            )

            if 1 == len(heuristic_result_list):
                heuristic_source = heuristic_result_list[0]
            else:
                if not skip_query:
                    print(f"{colors.ORANGE}Select search source{colors.END}:")
                    for i, heuristic_source in enumerate(heuristic_result_list):
                        highlight_color = ""
                        if heuristic_source["confidence"] >= 0.7:
                            highlight_color = colors.GREEN
                        elif heuristic_source["confidence"] >= 0.5:
                            highlight_color = colors.ORANGE
                        print(
                            f"{highlight_color}{i+1} "
                            f"(confidence: {round(heuristic_source['confidence'], 2)}):"
                            f" {heuristic_source['source_candidate'].endpoint}{colors.END}"
                        )

                while True:
                    if skip_query:
                        # Use the last / unknown_source
                        selection = str(len(heuristic_result_list))
                    else:
                        selection = input("select nr")
                    if not selection.isdigit():
                        continue
                    if int(selection) in range(1, len(heuristic_result_list) + 1):
                        heuristic_source = heuristic_result_list[int(selection) - 1]
                        break

            if "colrev.unknown_source" == heuristic_source["source_candidate"].endpoint:
                cmd = "Enter the search query (or NA)".ljust(25, " ") + ": "
                query_input = ""
                if not skip_query:
                    query_input = input(cmd)
                if query_input not in ["", "NA"]:
                    heuristic_source["source_candidate"].search_parameters = {
                        "query": query_input
                    }
                else:
                    heuristic_source["source_candidate"].search_parameters = {}

            self.review_manager.logger.info(
                f"Source name: {heuristic_source['source_candidate'].endpoint}"
            )

            heuristic_source["source_candidate"].comment = None

            if (
                {}
                == heuristic_source["source_candidate"].load_conversion_package_endpoint
            ):
                custom_load_conversion_package_endpoint = input(
                    "provide custom load_conversion_package_endpoint [or NA]:"
                )
                if "NA" == custom_load_conversion_package_endpoint:
                    heuristic_source[
                        "source_candidate"
                    ].load_conversion_package_endpoint = {}
                else:
                    heuristic_source[
                        "source_candidate"
                    ].load_conversion_package_endpoint = {
                        "endpoint": custom_load_conversion_package_endpoint
                    }

            new_sources.append(heuristic_source["source_candidate"])

        return new_sources

    def __check_bib_file(
        self, *, source: colrev.settings.SearchSource, records: dict
    ) -> None:
        if len(records.items()) <= 3:
            return
        if not any("author" in r for ID, r in records.items()):
            raise colrev_exceptions.ImportException(
                f"Import failed (no record with author field): {source.filename.name}"
            )

        if not any("title" in r for ID, r in records.items()):
            raise colrev_exceptions.ImportException(
                f"Import failed (no record with title field): {source.filename.name}"
            )

    def __resolve_non_unique_ids(self, *, source: colrev.settings.SearchSource) -> None:
        def get_unique_id(*, non_unique_id: str, id_list: list[str]) -> str:
            order = 0
            letters = list(string.ascii_lowercase)
            temp_id = non_unique_id
            next_unique_id = temp_id
            appends: list = []
            while next_unique_id in id_list:
                if len(appends) == 0:
                    order += 1
                    appends = list(itertools.product(letters, repeat=order))
                next_unique_id = temp_id + "".join(list(appends.pop(0)))

            return next_unique_id

        def inplace_change_second(
            *, filename: Path, old_string: str, new_string: str
        ) -> None:
            new_file_lines = []
            with open(filename, encoding="utf8") as file:
                first_read = False
                replaced = False
                for line in file.readlines():
                    if old_string in line and not first_read:
                        first_read = True
                    if old_string in line and first_read and not replaced:
                        line = line.replace(old_string, new_string)
                        replaced = True
                    new_file_lines.append(line)

                # s = f.read()
                # if old_string not in s:
                #     return
            with open(filename, "w", encoding="utf8") as file:
                for new_file_line in new_file_lines:
                    file.write(new_file_line)

        if not source.get_corresponding_bib_file().is_file():
            return

        with open(source.get_corresponding_bib_file(), encoding="utf8") as bibtex_file:
            cr_dict = self.review_manager.dataset.load_records_dict(
                load_str=bibtex_file.read()
            )

        ids_to_update = []
        current_ids = list(cr_dict.keys())
        for record in cr_dict.values():
            if len([x for x in current_ids if x == record["ID"]]) > 1:
                new_id = get_unique_id(non_unique_id=record["ID"], id_list=current_ids)
                ids_to_update.append([record["ID"], new_id])
                current_ids.append(new_id)

        if len(ids_to_update) > 0:
            self.review_manager.dataset.add_changes(
                path=source.get_corresponding_bib_file()
            )
            self.review_manager.create_commit(
                msg=f"Save original search file: {source.get_corresponding_bib_file().name}",
            )

            for old_id, new_id in ids_to_update:
                self.review_manager.logger.info(
                    f"Resolve ID to ensure unique colrev_origins: {old_id} -> {new_id}"
                )
                self.review_manager.report_logger.info(
                    f"Resolve ID to ensure unique colrev_origins: {old_id} -> {new_id}"
                )
                inplace_change_second(
                    filename=source.get_corresponding_bib_file(),
                    old_string=f"{old_id},",
                    new_string=f"{new_id},",
                )
            self.review_manager.dataset.add_changes(
                path=source.get_corresponding_bib_file()
            )
            self.review_manager.create_commit(
                msg=f"Resolve non-unique IDs in {source.get_corresponding_bib_file().name}"
            )

    def __getbib(self, *, file: Path) -> list[dict]:
        with open(file, encoding="utf8") as bibtex_file:
            contents = bibtex_file.read()
            bib_r = re.compile(r"@.*{.*,", re.M)
            if len(re.findall(bib_r, contents)) == 0:
                self.review_manager.logger.error(f"Not a bib file? {file.name}")

        with open(file, encoding="utf8") as bibtex_file:
            search_records_dict = self.review_manager.dataset.load_records_dict(
                load_str=bibtex_file.read()
            )
        return list(search_records_dict.values())

    def __unescape_latex(self, *, input_str: str) -> str:
        # Based on
        # https://en.wikibooks.org/wiki/LaTeX/Special_Characters

        for latex_char, repl_char in self.__LATEX_SPECIAL_CHAR_MAPPING.items():
            input_str = input_str.replace(latex_char, repl_char)

        input_str = input_str.replace("\\emph", "")
        input_str = input_str.replace("\\textit", "")

        return input_str

    def __unescape_html(self, *, input_str: str) -> str:
        input_str = html.unescape(input_str)
        if "<" in input_str:
            input_str = re.sub(r"<.*?>", "", input_str)
        return input_str

    def import_provenance(
        self,
        *,
        record: colrev.record.Record,
    ) -> None:
        """Set the provenance for an imported record"""

        def set_initial_import_provenance(*, record: colrev.record.Record) -> None:
            # Initialize colrev_masterdata_provenance
            colrev_masterdata_provenance, colrev_data_provenance = {}, {}

            for key in record.data.keys():
                if key in colrev.record.Record.identifying_field_keys:
                    if key not in colrev_masterdata_provenance:
                        colrev_masterdata_provenance[key] = {
                            "source": record.data["colrev_origin"][0],
                            "note": "",
                        }
                elif key not in colrev.record.Record.provenance_keys and key not in [
                    "colrev_source_identifier",
                    "ID",
                    "ENTRYTYPE",
                ]:
                    colrev_data_provenance[key] = {
                        "source": record.data["colrev_origin"][0],
                        "note": "",
                    }

            record.data["colrev_data_provenance"] = colrev_data_provenance
            record.data["colrev_masterdata_provenance"] = colrev_masterdata_provenance

        def set_initial_non_curated_import_provenance(
            *, record: colrev.record.Record
        ) -> None:
            masterdata_restrictions = (
                self.review_manager.dataset.get_applicable_restrictions(
                    record_dict=record.get_data()
                )
            )
            if masterdata_restrictions:
                record.update_masterdata_provenance(
                    masterdata_restrictions=masterdata_restrictions
                )
            else:
                record.apply_fields_keys_requirements()

            if (
                record.data["ENTRYTYPE"]
                in colrev.record.Record.record_field_inconsistencies
            ):
                inconsistent_fields = colrev.record.Record.record_field_inconsistencies[
                    record.data["ENTRYTYPE"]
                ]
                for inconsistent_field in inconsistent_fields:
                    if inconsistent_field in record.data:
                        inconsistency_note = (
                            f"inconsistent with entrytype ({record.data['ENTRYTYPE']})"
                        )
                        record.add_masterdata_provenance_note(
                            key=inconsistent_field, note=inconsistency_note
                        )

            incomplete_fields = record.get_incomplete_fields()
            for incomplete_field in incomplete_fields:
                record.add_masterdata_provenance_note(
                    key=incomplete_field, note="incomplete"
                )

            defect_fields = record.get_quality_defects()
            if defect_fields:
                for defect_field in defect_fields:
                    record.add_masterdata_provenance_note(
                        key=defect_field, note="quality_defect"
                    )

        if not record.masterdata_is_curated():
            set_initial_import_provenance(record=record)
            set_initial_non_curated_import_provenance(record=record)

    def __import_record(self, *, record_dict: dict) -> dict:
        # pylint: disable=too-many-branches

        self.review_manager.logger.debug(
            f'import_record {record_dict["ID"]}: '
            # f"\n{self.review_manager.p_printer.pformat(record_dict)}\n\n"
        )

        if colrev.record.RecordState.md_retrieved == record_dict["colrev_status"]:
            # Consistently set keys to lower case
            lower_keys = [k.lower() for k in list(record_dict.keys())]
            for key, n_key in zip(list(record_dict.keys()), lower_keys):
                if key in ["ID", "ENTRYTYPE"]:
                    continue
                record_dict[n_key] = record_dict.pop(key)

            # pylint: disable=duplicate-code
            # For better readability of the git diff:
            fields_to_process = [
                "author",
                "year",
                "title",
                "journal",
                "booktitle",
                "series",
                "volume",
                "number",
                "pages",
                "doi",
                "abstract",
            ]

            for field in fields_to_process:
                if field in record_dict:
                    if "\\" in record_dict[field]:
                        record_dict[field] = self.__unescape_latex(
                            input_str=record_dict[field]
                        )

                    if "<" in record_dict[field]:
                        record_dict[field] = self.__unescape_html(
                            input_str=record_dict[field]
                        )

                    record_dict[field] = (
                        record_dict[field]
                        .replace("\n", " ")
                        .rstrip()
                        .lstrip()
                        .replace("{", "")
                        .replace("}", "")
                    )

            if "UNKNOWN" != record_dict.get("title", "UNKNOWN"):
                record_dict["title"] = (
                    re.sub(r"\s+", " ", record_dict["title"])
                    .lstrip()
                    .rstrip()
                    .rstrip(".")
                )

            if "year" in record_dict:
                if str(record_dict["year"]).endswith(".0"):
                    record_dict["year"] = str(record_dict["year"])[:-2]

            if "pages" in record_dict:
                record_dict["pages"] = record_dict["pages"].replace("–", "--")
                if record_dict["pages"].count("-") == 1:
                    record_dict["pages"] = record_dict["pages"].replace("-", "--")
                if "n.pag" == record_dict["pages"].lower():
                    del record_dict["pages"]

            if "number" not in record_dict and "issue" in record_dict:
                record_dict.update(number=record_dict["issue"])
                del record_dict["issue"]

            if record_dict.get("volume", "") == "ahead-of-print":
                del record_dict["volume"]
            if record_dict.get("number", "") == "ahead-of-print":
                del record_dict["number"]

            if "language" in record_dict:
                if len(record_dict["language"]) > 3:
                    for long, short in [
                        ("English", "eng"),
                        ("Spanish", "esp"),
                        ("Chinese", "zho"),
                        ("Portuguese", "por"),
                        ("German", "deu"),
                        ("Hungarian", "hun"),
                        ("French", "fra"),
                        ("Russian", "rus"),
                    ]:
                        record_dict["language"] = record_dict["language"].replace(
                            long, short
                        )

            if "url" in record_dict:
                if "login?url=https" in record_dict["url"]:
                    record_dict["url"] = record_dict["url"][
                        record_dict["url"].find("login?url=https") + 10 :
                    ]

        record = colrev.record.Record(data=record_dict)
        if "doi" in record.data:
            record.data.update(
                doi=record.data["doi"].replace("http://dx.doi.org/", "").upper()
            )

        self.import_provenance(
            record=record,
        )
        record.set_status(target_state=colrev.record.RecordState.md_imported)

        if record.check_potential_retracts():
            self.review_manager.logger.info(
                f"{colors.GREEN}Found paper retract: "
                f"{record.data['ID']}{colors.END}"
            )

        return record.get_data()

    def __prep_records_for_import(
        self, *, source: colrev.settings.SearchSource, search_records: list
    ) -> list:
        record_list = []
        origin_prefix = source.get_origin_prefix()
        for record in search_records:
            for key in colrev.record.Record.provenance_keys + [
                "screening_criteria",
            ]:
                if "colrev_status" == key:
                    continue
                if key in record:
                    del record[key]

            record.update(colrev_origin=[f"{origin_prefix}/{record['ID']}"])

            # Drop empty fields
            record = {k: v for k, v in record.items() if v}

            post_md_prepared_states = colrev.record.RecordState.get_post_x_states(
                state=colrev.record.RecordState.md_prepared
            )

            if record.get("colrev_status", "") in post_md_prepared_states:
                # Note : when importing a record, it always needs to be
                # deduplicated against the other records in the repository
                record.update(colrev_status=colrev.record.RecordState.md_prepared)
            else:
                record.update(colrev_status=colrev.record.RecordState.md_retrieved)

            if "doi" in record:
                formatted_doi = (
                    record["doi"]
                    .lower()
                    .replace("https://", "http://")
                    .replace("dx.doi.org", "doi.org")
                    .replace("http://doi.org/", "")
                    .upper()
                )
                record.update(doi=formatted_doi)
                # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
                doi_match = re.match(r"^10.\d{4,9}\/", record["doi"])
                if not doi_match:
                    self.review_manager.logger.info(
                        f"remove doi (not matching regex): {record['doi']}"
                    )
                    del record["doi"]

            self.review_manager.logger.debug(
                f'append record {record["ID"]} '
                # f"\n{self.review_manager.p_printer.pformat(record)}\n\n"
            )
            record_list.append(record)
        return record_list

    def __get_search_records(self, *, source: colrev.settings.SearchSource) -> list:
        search_records = []
        if source.get_corresponding_bib_file().is_file():
            search_records = self.__getbib(file=source.get_corresponding_bib_file())
            self.review_manager.logger.debug(
                f"Loaded {source.get_corresponding_bib_file().name} "
                f"with {len(search_records)} records"
            )
        else:
            self.review_manager.logger.error(
                f"Did not find bib file {source.get_corresponding_bib_file().name} "
            )
            return []

        if len(search_records) == 0:
            self.review_manager.logger.info(
                f"{colors.GREEN}No records to load{colors.END}"
            )
            return search_records

        nr_in_bib = self.review_manager.dataset.get_nr_in_bib(
            file_path=source.get_corresponding_bib_file()
        )
        if len(search_records) < nr_in_bib:
            self.review_manager.logger.error(
                "broken bib file (not imported all records)"
            )
            with open(source.get_corresponding_bib_file(), encoding="utf8") as file:
                line = file.readline()
                while line:
                    if "@" in line[:3]:
                        record_id = line[line.find("{") + 1 : line.rfind(",")]
                        if record_id not in [x["ID"] for x in search_records]:
                            self.review_manager.logger.error(
                                f"{record_id} not imported"
                            )
                    line = file.readline()
        return search_records

    def __load_source_records(
        self, *, source: colrev.settings.SearchSource, keep_ids: bool
    ) -> None:
        search_records = self.__get_search_records(source=source)

        record_list = self.__prep_records_for_import(
            source=source, search_records=search_records
        )

        imported_origins = self.__get_currently_imported_origin_list()
        record_list = [
            x for x in record_list if x["colrev_origin"][0] not in imported_origins
        ]
        source.setup_for_load(
            record_list=record_list, imported_origins=imported_origins
        )
        if len(search_records) == 0:
            return

        records = self.review_manager.dataset.load_records_dict()
        for source_record in source.source_records_list:
            source_record = self.__import_record(record_dict=source_record)

            # Make sure IDs are unique / do not replace existing records
            order = 0
            letters = list(string.ascii_lowercase)
            next_unique_id = source_record["ID"]
            appends: list = []
            while next_unique_id in records:
                if len(appends) == 0:
                    order += 1
                    appends = list(itertools.product(letters, repeat=order))
                next_unique_id = source_record["ID"] + "".join(list(appends.pop(0)))
            source_record["ID"] = next_unique_id
            records[source_record["ID"]] = source_record

            self.review_manager.logger.info(
                f" {colors.GREEN}{source_record['ID']}".ljust(46)
                + f"md_retrieved →  md_imported{colors.END}"
            )

        self.__check_bib_file(source=source, records=records)
        self.review_manager.dataset.save_records_dict(records=records)

        if keep_ids:
            self.review_manager.logger.warning(
                "Not yet fully implemented. Need to check/resolve ID duplicates."
            )
        else:
            self.review_manager.logger.debug("Set IDs")
            records = self.review_manager.dataset.set_ids(
                records=records,
                selected_ids=[r["ID"] for r in source.source_records_list],
            )

        self.review_manager.logger.info(
            "New records loaded".ljust(38) + f"{source.to_import} records"
        )

        self.review_manager.dataset.add_setting_changes()
        self.review_manager.dataset.add_changes(
            path=source.get_corresponding_bib_file()
        )
        self.review_manager.dataset.add_changes(path=source.filename)
        self.review_manager.dataset.add_record_changes()

    def __validate_load(self, *, source: colrev.settings.SearchSource) -> None:
        imported_origins = self.__get_currently_imported_origin_list()
        imported = len(imported_origins) - source.len_before

        if imported != source.to_import:
            # Note : for diagnostics, it is easier if we complete the process
            # and create the commit (instead of raising an exception)
            self.review_manager.logger.error(f"len_before: {source.len_before}")
            self.review_manager.logger.error(f"len_after: {len(imported_origins)}")

            origins_to_import = [o["colrev_origin"] for o in source.source_records_list]
            if source.to_import - imported > 0:
                self.review_manager.logger.error(
                    f"{colors.RED}PROBLEM: delta: "
                    f"{source.to_import - imported} records missing{colors.END}"
                )

                missing_origins = [
                    o for o in origins_to_import if o not in imported_origins
                ]
                self.review_manager.logger.error(
                    f"{colors.RED}Records not yet imported: {missing_origins}{colors.END}"
                )
            else:
                self.review_manager.logger.error(
                    f"{colors.RED}PROBLEM: "
                    f"{source.to_import - imported} records too much{colors.END}"
                )

    def __save_records(self, *, records: dict, corresponding_bib_file: Path) -> None:
        def fix_keys(*, records: dict) -> dict:
            for record in records.values():
                record = {
                    re.sub("[0-9a-zA-Z_]+", "1", k.replace(" ", "_")): v
                    for k, v in record.items()
                }
            return records

        def set_incremental_ids(*, records: dict) -> dict:
            # if IDs to set for some records
            if 0 != len([r for r in records if "ID" not in r]):
                i = 1
                for record in records.values():
                    if "ID" not in record:
                        if "UT_(Unique_WOS_ID)" in record:
                            record["ID"] = record["UT_(Unique_WOS_ID)"].replace(
                                ":", "_"
                            )
                        else:
                            record["ID"] = f"{i+1}".rjust(10, "0")
                        i += 1
            return records

        def drop_empty_fields(*, records: dict) -> dict:
            for record_id in records:
                records[record_id] = {
                    k: v for k, v in records[record_id].items() if v is not None
                }
                records[record_id] = {
                    k: v for k, v in records[record_id].items() if v != "nan"
                }
            return records

        def drop_fields(*, records: dict) -> dict:
            for record_id in records:
                records[record_id] = {
                    k: v
                    for k, v in records[record_id].items()
                    if k not in ["colrev_status", "colrev_masterdata_provenance"]
                }
            return records

        if len(records) == 0:
            self.review_manager.report_logger.debug("No records loaded")
            self.review_manager.logger.debug("No records loaded")
            return

        records = fix_keys(records=records)
        records = set_incremental_ids(records=records)
        records = drop_empty_fields(records=records)
        records = drop_fields(records=records)

        self.review_manager.dataset.save_records_dict_to_file(
            records=records, save_path=corresponding_bib_file
        )

    def __load_active_sources(
        self, *, new_sources: typing.List[colrev.settings.SearchSource]
    ) -> list:
        checker = self.review_manager.get_checker()
        checker.check_sources()
        sources = []
        for source in self.review_manager.settings.sources:
            sources.append(source)
        for source in new_sources:
            if source.filename not in [s.filename for s in sources]:
                sources.append(source)
        return sources

    def main(
        self,
        *,
        new_sources: typing.List[colrev.settings.SearchSource],
        keep_ids: bool = False,
        combine_commits: bool = False,
    ) -> None:
        """Load records (main entrypoint)"""

        if not self.review_manager.high_level_operation:
            print()
        git_repo = self.review_manager.dataset.get_repo()
        part_exact_call = self.review_manager.exact_call
        for source in self.__load_active_sources(new_sources=new_sources):
            try:
                self.review_manager.logger.info(f"Load {source.filename}")

                # Add to settings (if new filename)
                if source.filename not in [
                    s.filename for s in self.review_manager.settings.sources
                ]:
                    self.review_manager.settings.sources.append(source)
                    self.review_manager.save_settings()
                    # Add files that were renamed (removed)
                    for obj in git_repo.index.diff(None).iter_change_type("D"):
                        if source.filename.stem in obj.b_path:
                            self.review_manager.dataset.add_changes(
                                path=Path(obj.b_path), remove=True
                            )

                # 1. convert to bib and fix format (if necessary)
                load_conversion_package_endpoint_dict = self.package_manager.load_packages(
                    package_type=colrev.env.package_manager.PackageEndpointType.load_conversion,
                    selected_packages=[source.load_conversion_package_endpoint],
                    operation=self,
                )

                load_conversion_package_endpoint = (
                    load_conversion_package_endpoint_dict[
                        source.load_conversion_package_endpoint["endpoint"]
                    ]
                )
                records = load_conversion_package_endpoint.load(self, source)  # type: ignore
                self.__save_records(
                    records=records,
                    corresponding_bib_file=source.get_corresponding_bib_file(),
                )

                # 2. resolve non-unique IDs (if any)
                self.__resolve_non_unique_ids(source=source)

                # 3. load and add records to data/records.bib
                self.__load_source_records(source=source, keep_ids=keep_ids)
                if (
                    0 == getattr(source, "to_import", 0)
                    and not self.review_manager.high_level_operation
                ):
                    print()

                # 4. validate load
                self.__validate_load(source=source)

                stashed = "No local changes to save" != git_repo.git.stash(
                    "push", "--keep-index"
                )

                if not combine_commits:
                    self.review_manager.exact_call = (
                        f"{part_exact_call} -s {source.filename.name}"
                    )
                    self.review_manager.create_commit(
                        msg=f"Load {source.filename.name}",
                    )
                if stashed:
                    git_repo.git.stash("pop")
                if not self.review_manager.high_level_operation:
                    print()
            except colrev_exceptions.ImportException as exc:
                print(exc)

        if combine_commits and self.review_manager.dataset.has_changes():
            self.review_manager.create_commit(msg="Load (multiple)")

        self.review_manager.logger.info(
            f"{colors.GREEN}Completed load operation{colors.END}"
        )
        if self.review_manager.in_ci_environment():
            print("\n\n")


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Completion of metadata based on year-volume-issue dependency as a prep operation"""
from __future__ import annotations

from dataclasses import dataclass

import requests
import timeout_decorator
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.built_in.search_sources.crossref as crossref_connector
import colrev.ops.search_sources
import colrev.record

# pylint: disable=duplicate-code
if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.prep

# pylint: disable=too-few-public-methods


@zope.interface.implementer(colrev.env.package_manager.PrepPackageEndpointInterface)
@dataclass
class YearVolIssPrep(JsonSchemaMixin):
    """Prepares records based on year-volume-issue dependency"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = True

    source_correction_hint = (
        "ask the publisher to correct the metadata"
        + " (see https://www.crossref.org/blog/"
        + "metadata-corrections-updates-and-additions-in-metadata-manager/"
    )
    always_apply_changes = True

    def __init__(
        self,
        *,
        prep_operation: colrev.ops.prep.Prep,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        self.settings = self.settings_class.load_settings(data=settings)

        self.local_index = prep_operation.review_manager.get_local_index()

        vol_nr_dict: dict = {}
        if hasattr(prep_operation.review_manager, "dataset"):
            records = prep_operation.review_manager.dataset.load_records_dict()
            for record in records.values():
                # pylint: disable=duplicate-code
                if record[
                    "colrev_status"
                ] not in colrev.record.RecordState.get_post_x_states(
                    state=colrev.record.RecordState.md_processed
                ):
                    continue
                if not record.get("year", "NA").isdigit():
                    continue

                if "journal" not in record or "volume" not in record:
                    continue

                if record["journal"] not in vol_nr_dict:
                    vol_nr_dict[record["journal"]] = {}

                if record["volume"] not in vol_nr_dict[record["journal"]]:
                    vol_nr_dict[record["journal"]][record["volume"]] = {}

                if "number" not in record:
                    vol_nr_dict[record["journal"]][record["volume"]] = record["year"]
                else:
                    if isinstance(
                        vol_nr_dict[record["journal"]][record["volume"]], dict
                    ):
                        vol_nr_dict[record["journal"]][record["volume"]][
                            record["number"]
                        ] = record["year"]
                    else:
                        # do not use inconsistent data (has/has no number)
                        del vol_nr_dict[record["journal"]][record["volume"]]

        self.vol_nr_dict = vol_nr_dict

    @timeout_decorator.timeout(60, use_signals=False)
    def prepare(
        self, prep_operation: colrev.ops.prep.Prep, record: colrev.record.PrepRecord
    ) -> colrev.record.Record:
        """Prepare a record based on year-volume-issue dependency"""

        # pylint: disable=too-many-nested-blocks
        # pylint: disable=too-many-return-statements
        # pylint: disable=too-many-branches

        if record.data.get("year", "NA").isdigit() or record.masterdata_is_curated():
            return record

        # TBD: maybe extract the following three lines as a separate script...
        try:
            year = self.local_index.get_year_from_toc(record_dict=record.get_data())
            record.update_field(
                key="year",
                value=year,
                source="LocalIndexPrep",
                note="",
                keep_source_if_equal=True,
            )
            return record
        except colrev_exceptions.TOCNotAvailableException:
            pass

        if "journal" in record.data:
            if record.data["journal"] in self.vol_nr_dict:
                if "volume" in record.data:
                    if (
                        record.data["volume"]
                        in self.vol_nr_dict[record.data["journal"]]
                    ):
                        if "number" in record.data:
                            if (
                                record.data["number"]
                                in self.vol_nr_dict[record.data["journal"]][
                                    record.data["volume"]
                                ]
                            ):
                                record.update_field(
                                    key="year",
                                    value=self.vol_nr_dict[record.data["journal"]][
                                        record.data["volume"]
                                    ][record.data["number"]],
                                    source="year_vol_iss_prep",
                                    note="",
                                )
                                record.update_masterdata_provenance()
                                return record
                        else:
                            if isinstance(
                                self.vol_nr_dict[record.data["journal"]][
                                    record.data["volume"]
                                ],
                                (str, int),
                            ):
                                record.update_field(
                                    key="year",
                                    value=self.vol_nr_dict[record.data["journal"]][
                                        record.data["volume"]
                                    ],
                                    source="year_vol_iss_prep",
                                    note="",
                                )
                                record.update_masterdata_provenance()
                                return record

        # The year depends on journal x volume x issue
        if (
            "journal" in record.data
            and "volume" in record.data
            and "number" in record.data
        ) and "UNKNOWN" == record.data.get("year", "UNKNOWN"):
            pass
        else:
            return record

        crossref_source = crossref_connector.CrossrefSearchSource(
            source_operation=prep_operation
        )
        try:
            retrieved_records = crossref_source.crossref_query(
                review_manager=prep_operation.review_manager,
                record_input=record,
                jour_vol_iss_list=True,
                timeout=prep_operation.timeout,
            )
            retries = 0
            while (
                not retrieved_records and retries < prep_operation.max_retries_on_error
            ):
                retries += 1
                retrieved_records = crossref_source.crossref_query(
                    review_manager=prep_operation.review_manager,
                    record_input=record,
                    jour_vol_iss_list=True,
                    timeout=prep_operation.timeout,
                )
            if 0 == len(retrieved_records):
                return record

            retrieved_records = [
                retrieved_record
                for retrieved_record in retrieved_records
                if retrieved_record.data.get("volume", "NA")
                == record.data.get("volume", "NA")
                and retrieved_record.data.get("journal", "NA")
                == record.data.get("journal", "NA")
                and retrieved_record.data.get("number", "NA")
                == record.data.get("number", "NA")
            ]

            years = [r.data["year"] for r in retrieved_records]
            if len(years) == 0:
                return record
            most_common = max(years, key=years.count)
            if years.count(most_common) > 3:
                record.update_field(
                    key="year", value=most_common, source="CROSSREF(average)", note=""
                )
        except requests.exceptions.RequestException:
            pass

        return record


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
"""Scripts to add packages using the cli."""
from __future__ import annotations

from pathlib import Path

import requests

import colrev.env.package_manager
import colrev.ops.built_in.data.bibliography_export
import colrev.ui_cli.cli_colors as colors


def add_search_source(
    *,
    search_operation: colrev.ops.search.Search,
    query: str,
) -> None:
    """Add a search source package_endpoint"""

    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches

    if Path(query).is_file():
        # pylint: disable=import-outside-toplevel
        import shutil

        filename = search_operation.get_unique_filename(
            file_path_string=Path(query).name
        )
        dst = search_operation.review_manager.search_dir / Path(filename).name
        shutil.copyfile(query, dst)
        search_operation.review_manager.logger.info(f"Copied {filename} to repo.")
        load_operation = search_operation.review_manager.get_load_operation()
        new_sources = load_operation.get_new_sources()
        load_operation.main(new_sources=new_sources)
        # Note : load runs the heuristics.
        return

    if ":" not in query:
        search_operation.review_manager.logger.error(
            "Could not find package identifier at the beginning of the query"
        )
        return
    package_identifier = query[: query.find(":")]
    query = query[query.find(":") + 1 :]

    package_manager = search_operation.review_manager.get_package_manager()

    search_source = package_manager.load_packages(
        package_type=colrev.env.package_manager.PackageEndpointType.search_source,
        selected_packages=[{"endpoint": package_identifier}],
        operation=search_operation,
        instantiate_objects=False,
    )

    add_source = search_source[package_identifier].add_endpoint(search_operation, query)  # type: ignore
    search_operation.add_source(add_source=add_source)


def add_data(
    *,
    data_operation: colrev.ops.data.Data,
    add: str,
) -> None:
    """Add a data package_endpoint"""

    package_manager = data_operation.review_manager.get_package_manager()
    available_data_endpoints = package_manager.discover_packages(
        package_type=colrev.env.package_manager.PackageEndpointType.data
    )
    data_operation.review_manager.logger.info(f"Add {add} data endpoint")
    if add in available_data_endpoints:
        package_endpoints = package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.data,
            selected_packages=[{"endpoint": add}],
            operation=data_operation,
        )
        endpoint = package_endpoints[add]

        default_endpoint_conf = endpoint.get_default_setup()  # type: ignore

        if "colrev.paper_md" == add:
            if "y" == input("Select a custom word template (y/n)?"):
                template_name = input(
                    'Please copy the word template to " \
                "the project directory and enter the filename.'
                )
                default_endpoint_conf["word_template"] = template_name
            else:
                print("Adding APA as a default")

            if "y" == input("Select a custom citation stlye (y/n)?"):
                print(
                    "Citation stlyes are available at: \n"
                    "https://github.com/citation-style-language/styles"
                )
                csl_link = input("Please select a citation style and provide the link.")
                ret = requests.get(csl_link, allow_redirects=True, timeout=30)
                with open(Path(csl_link).name, "wb") as file:
                    file.write(ret.content)
            else:
                print("Adding APA as a default")

            data_operation.review_manager.dataset.add_changes(
                path=Path(Path(csl_link).name)
            )
            data_operation.review_manager.dataset.add_changes(
                path=default_endpoint_conf["word_template"]
            )

        data_operation.add_data_endpoint(data_endpoint=default_endpoint_conf)
        data_operation.review_manager.create_commit(
            msg="Add data endpoint",
            script_call="colrev data",
        )

    elif add in [
        e.value for e in colrev.ops.built_in.data.bibliography_export.BibFormats
    ]:
        package_endpoints = package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.data,
            selected_packages=[{"endpoint": "colrev.bibliography_export"}],
            operation=data_operation,
        )
        endpoint = package_endpoints["colrev.bibliography_export"]
        default_endpoint_conf = endpoint.get_default_setup()  # type: ignore
        default_endpoint_conf["bib_format"] = add
        data_operation.add_data_endpoint(data_endpoint=default_endpoint_conf)
        data_operation.review_manager.create_commit(
            msg=f"Add {add} data endpoint",
            script_call="colrev data",
        )
    else:
        print("Data format not available")
        return

    # Note : reload updated settings
    review_manager = colrev.review_manager.ReviewManager(force_mode=True)
    data_operation = colrev.ops.data.Data(review_manager=review_manager)

    data_operation.main(selection_list=["colrev.bibliography_export"], silent_mode=True)
    data_operation.review_manager.logger.info(
        f"{colors.GREEN}Successfully added {add} data endpoint{colors.END}"
    )


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""SearchSource: ABI/INFORM (ProQuest)"""
from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class ABIInformProQuestSearchSource(JsonSchemaMixin):
    """SearchSource for ABI/INFORM (ProQuest)"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "{{ID}}"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "ABI/INFORM (ProQuest)"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/colrev/"
        + "ops/built_in/search_sources/abi_inform_proquest.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.CheckOperation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for ABI/INFORM (ProQuest)"""

        result = {"confidence": 0.0}

        if data.count("\n@") > 1:
            if data.count("www.proquest.com") >= data.count("\n@"):
                result["confidence"] = 1.0

        return result

    def validate_source(
        self,
        search_operation: colrev.ops.search.Search,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Validate the SearchSource (parameters etc.)"""

    @classmethod
    def add_endpoint(
        cls, search_operation: colrev.ops.search.Search, query: str
    ) -> typing.Optional[colrev.settings.SearchSource]:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        return None

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of ABI/INFORM"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load_fixes(
        self,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
        records: typing.Dict,
    ) -> dict:
        """Load fixes for ABI/INFORM (ProQuest)"""

        to_delete = []
        for record in records.values():
            if re.search(r"-\d{1,2}$", record["ID"]):
                original_record_id = re.sub(r"-\d{1,2}$", "", record["ID"])
                original_record = records[original_record_id]

                # Note: between duplicate records, there are variations in spelling and completeness
                if (
                    colrev.record.Record.get_record_similarity(
                        record_a=colrev.record.Record(data=record),
                        record_b=colrev.record.Record(data=original_record),
                    )
                    < 0.9
                ):
                    continue

                if original_record_id not in records:
                    continue
                to_delete.append(record["ID"])

        for rid in to_delete:
            load_operation.review_manager.logger.info(f" remove duplicate {rid}")
            del records[rid]

        load_operation.review_manager.dataset.save_records_dict_to_file(
            records=records, save_path=source.filename
        )

        return records

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for ABI/INFORM (ProQuest)"""

        if "language" in record.data:
            if record.data["language"] in ["ENG", "English"]:
                record.update_field(
                    key="language",
                    value="eng",
                    source="prep_abi_inform_proquest_source",
                )

        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Prescreen based on specified scope"""
from __future__ import annotations

import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.language_service
import colrev.env.package_manager
import colrev.record

if False:  # pylint: disable=using-constant-test
    if typing.TYPE_CHECKING:
        import colrev.ops.prescreen.Prescreen

# pylint: disable=too-few-public-methods
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.PrescreenPackageEndpointInterface
)
@dataclass
class ScopePrescreen(JsonSchemaMixin):

    """Rule-based prescreen (scope)"""

    ci_supported: bool = True

    @dataclass
    class ScopePrescreenSettings(
        colrev.env.package_manager.DefaultSettings, JsonSchemaMixin
    ):
        """Settings for ScopePrescreen"""

        # pylint: disable=invalid-name
        # pylint: disable=too-many-instance-attributes

        endpoint: str
        TimeScopeFrom: typing.Optional[int]
        TimeScopeTo: typing.Optional[int]
        LanguageScope: typing.Optional[list]
        ExcludeComplementaryMaterials: typing.Optional[bool]
        OutletInclusionScope: typing.Optional[dict]
        OutletExclusionScope: typing.Optional[dict]
        ENTRYTYPEScope: typing.Optional[list]

        _details = {
            "TimeScopeFrom": {
                "tooltip": "Lower bound for the time scope",
                "min": 1900,
                "max": 2050,
            },
            "TimeScopeTo": {
                "tooltip": "Upper bound for the time scope",
                "min": 1900,
                "max": 2050,
            },
            "LanguageScope": {"tooltip": "Language scope"},
            "ExcludeComplementaryMaterials": {
                "tooltip": "Whether complementary materials (coverpages etc.) are excluded"
            },
            "OutletInclusionScope": {
                "tooltip": "Particular outlets that should be included (exclusively)"
            },
            "OutletExclusionScope": {
                "tooltip": "Particular outlets that should be excluded"
            },
            "ENTRYTYPEScope": {
                "tooltip": "Particular ENTRYTYPEs that should be included (exclusively)"
            },
        }

    settings_class = ScopePrescreenSettings

    def __init__(
        self,
        *,
        prescreen_operation: colrev.ops.prescreen.Prescreen,  # pylint: disable=unused-argument
        settings: dict,
    ) -> None:
        if "TimeScopeFrom" in settings:
            assert settings["TimeScopeFrom"] > 1900
        if "TimeScopeFrom" in settings:
            assert settings["TimeScopeFrom"] < 2100
        if "TimeScopeTo" in settings:
            assert settings["TimeScopeTo"] > 1900
        if "TimeScopeTo" in settings:
            assert settings["TimeScopeTo"] < 2100
        if "LanguageScope" in settings:
            self.language_service = colrev.env.language_service.LanguageService()
            self.language_service.validate_iso_639_3_language_codes(
                lang_code_list=settings["LanguageScope"]
            )

        self.settings = self.settings_class.load_settings(data=settings)

        self.predatory_journals_beal = self.__load_predatory_journals_beal()

        self.title_complementary_materials_keywords = (
            colrev.env.utils.load_complementary_material_keywords()
        )

    def __load_predatory_journals_beal(self) -> dict:
        predatory_journals = {}

        filedata = colrev.env.utils.get_package_file_content(
            file_path=Path("template/ops/predatory_journals_beall.csv")
        )

        if filedata:
            for pred_journal in filedata.decode("utf-8").splitlines():
                predatory_journals[pred_journal.lower()] = pred_journal.lower()

        return predatory_journals

    def __conditional_prescreen(
        self, *, prescreen_operation: colrev.ops.prescreen.Prescreen, record_dict: dict
    ) -> None:
        # pylint: disable=too-many-branches
        # pylint: disable=too-many-nested-blocks

        if record_dict["colrev_status"] != colrev.record.RecordState.md_processed:
            return

        pad = 50

        # Note : LanguageScope is covered in prep
        # because dedupe cannot handle merges between languages

        if self.settings.ENTRYTYPEScope:
            if record_dict["ENTRYTYPE"] not in self.settings.ENTRYTYPEScope:
                colrev.record.Record(data=record_dict).prescreen_exclude(
                    reason="not in ENTRYTYPEScope"
                )

        if self.settings.OutletExclusionScope:
            if "values" in self.settings.OutletExclusionScope:
                for resource in self.settings.OutletExclusionScope["values"]:
                    for key, value in resource.items():
                        if key in record_dict and record_dict.get(key, "") == value:
                            colrev.record.Record(data=record_dict).prescreen_exclude(
                                reason="in OutletExclusionScope"
                            )
            if "list" in self.settings.OutletExclusionScope:
                for resource in self.settings.OutletExclusionScope["list"]:
                    for key, value in resource.items():
                        if "resource" == key and "predatory_journals_beal" == value:
                            if "journal" in record_dict:
                                if (
                                    record_dict["journal"].lower()
                                    in self.predatory_journals_beal
                                ):
                                    colrev.record.Record(
                                        data=record_dict
                                    ).prescreen_exclude(
                                        reason="predatory_journals_beal"
                                    )

        if self.settings.TimeScopeFrom:
            if int(record_dict.get("year", 0)) < self.settings.TimeScopeFrom:
                colrev.record.Record(data=record_dict).prescreen_exclude(
                    reason="not in TimeScopeFrom " f"(>{self.settings.TimeScopeFrom})"
                )

        if self.settings.TimeScopeTo:
            if int(record_dict.get("year", 5000)) > self.settings.TimeScopeTo:
                colrev.record.Record(data=record_dict).prescreen_exclude(
                    reason="not in TimeScopeTo " f"(<{self.settings.TimeScopeTo})"
                )

        if self.settings.OutletInclusionScope:
            in_outlet_scope = False
            if "values" in self.settings.OutletInclusionScope:
                for outlet in self.settings.OutletInclusionScope["values"]:
                    for key, value in outlet.items():
                        if key in record_dict and record_dict.get(key, "") == value:
                            in_outlet_scope = True
            if not in_outlet_scope:
                colrev.record.Record(data=record_dict).prescreen_exclude(
                    reason="not in OutletInclusionScope"
                )

        if self.settings.ExcludeComplementaryMaterials:
            if self.settings.ExcludeComplementaryMaterials:
                if "title" in record_dict:
                    if (
                        record_dict["title"].lower()
                        in self.title_complementary_materials_keywords
                    ):
                        colrev.record.Record(data=record_dict).prescreen_exclude(
                            reason="complementary material"
                        )

        if (
            record_dict["colrev_status"]
            == colrev.record.RecordState.rev_prescreen_excluded
        ):
            prescreen_operation.review_manager.report_logger.info(
                f' {record_dict["ID"]}'.ljust(pad, " ")
                + "Prescreen excluded (automatically)"
            )

    def run_prescreen(
        self,
        prescreen_operation: colrev.ops.prescreen.Prescreen,
        records: dict,
        split: list,  # pylint: disable=unused-argument
    ) -> dict:
        """Prescreen records based on the scope parameters"""

        for record_dict in records.values():
            self.__conditional_prescreen(
                prescreen_operation=prescreen_operation, record_dict=record_dict
            )

        prescreen_operation.review_manager.dataset.save_records_dict(records=records)
        prescreen_operation.review_manager.dataset.add_record_changes()
        prescreen_operation.review_manager.create_commit(
            msg="Pre-screen (scope)",
            manual_author=False,
        )
        return records


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""SearchSource: systematic-review-datasets"""
from __future__ import annotations

import typing
from dataclasses import dataclass
from pathlib import Path

import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.ops.built_in.search_sources.crossref
import colrev.ops.search
import colrev.record

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class SystematicReviewDatasetsSearchSource(JsonSchemaMixin):
    """SearchSource for systematic-review-datasets

    https://github.com/asreview/systematic-review-datasets

    csv files available in the releases (output directories) are supported:
    https://github.com/asreview/systematic-review-datasets/tags

    csv files containing article IDs are supported by source-specific preparation.
    metadata is added based on dois and pubmedid (openalex is not yet supported)

    """

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "{{ID}}"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = False
    ci_supported: bool = False
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "systematic-review-datasets"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/systematic_review_datasets.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.CheckOperation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.crossref_connector = (
            colrev.ops.built_in.search_sources.crossref.CrossrefSearchSource(
                source_operation=source_operation
            )
        )

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for systematic-review-datasets"""

        result = {"confidence": 0.0}

        if "doi,pmid,openalex_id,label_included" in data:
            result["confidence"] = 1.0

        if '"doi","pmid","openalex_id","label_included"' in data:
            result["confidence"] = 1.0

        if (
            "record_id,pubmedID,title,authors,abstract,year,label_included,label_abstract_screening"
            in data
        ):
            result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls, search_operation: colrev.ops.search.Search, query: str
    ) -> typing.Optional[colrev.settings.SearchSource]:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""
        return None

    def validate_source(
        self,
        search_operation: colrev.ops.search.Search,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Validate the SearchSource (parameters etc.)"""

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of SystematicReviewDatasets"""

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load_fixes(
        self,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
        records: typing.Dict,
    ) -> dict:
        """Load fixes for systematic-review-datasets"""

        for record in records.values():
            if "pmid" in record:
                record["pubmedid"] = record["pmid"].replace(
                    "https://pubmed.ncbi.nlm.nih.gov/", ""
                )
                del record["pmid"]

        return records

    def prepare(
        self, record: colrev.record.Record, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for systematic-review-datasets"""

        if "doi" in record.data:
            retrieved_record = self.crossref_connector.query_doi(doi=record.data["doi"])
            record.change_entrytype(new_entrytype=retrieved_record.data["ENTRYTYPE"])
            for key in [
                "journal",
                "booktitle",
                "volume",
                "number",
                "year",
                "pages",
                "author",
                "title",
            ]:
                if key in retrieved_record.data:
                    record.data[key] = retrieved_record.data[key]

        return record


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""Load conversion based on zotero importers (ris, rdf, json, mods, ...)"""
from __future__ import annotations

import json
from dataclasses import asdict
from dataclasses import dataclass

import requests
import zope.interface
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.ops.load

# pylint: disable=too-few-public-methods
# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.LoadConversionPackageEndpointInterface
)
@dataclass
class ZoteroTranslationLoader(JsonSchemaMixin):

    """Loads bibliography files (based on Zotero).
    Supports ris, rdf, json, mods, xml, marc, txt"""

    settings_class = colrev.env.package_manager.DefaultSettings
    ci_supported: bool = False

    supported_extensions = ["ris", "rdf", "json", "mods", "xml", "marc", "txt"]

    def __init__(self, *, load_operation: colrev.ops.load.Load, settings: dict):
        self.settings = self.settings_class.load_settings(data=settings)

        if not load_operation.review_manager.in_ci_environment():
            self.zotero_translation_service = (
                load_operation.review_manager.get_zotero_translation_service()
            )

    def load(
        self, load_operation: colrev.ops.load.Load, source: colrev.settings.SearchSource
    ) -> dict:
        """Load records from the source"""

        load_operation.review_manager.logger.info(
            "Starting Zotero translation services (Docker)"
        )
        self.zotero_translation_service.start()

        # pylint: disable=consider-using-with

        files = {"file": open(source.filename, "rb")}
        headers = {"Content-type": "text/plain"}
        ret = requests.post(
            "http://127.0.0.1:1969/import", headers=headers, files=files, timeout=30
        )

        headers = {"Content-type": "application/json"}
        if "No suitable translators found" == ret.content.decode("utf-8"):
            raise colrev_exceptions.ImportException(
                "Zotero translators: No suitable import translators found"
            )

        try:
            zotero_format = json.loads(ret.content)
            ret = requests.post(
                "http://127.0.0.1:1969/export?format=bibtex",
                headers=headers,
                json=zotero_format,
                timeout=60,
            )
            records = load_operation.review_manager.dataset.load_records_dict(
                load_str=ret.content.decode("utf-8")
            )

        except Exception as exc:
            raise colrev_exceptions.ImportException(
                f"Zotero import translators failed ({exc})"
            )

        self.zotero_translation_service.stop()

        endpoint_dict = load_operation.package_manager.load_packages(
            package_type=colrev.env.package_manager.PackageEndpointType.search_source,
            selected_packages=[asdict(source)],
            operation=load_operation,
            ignore_not_available=False,
        )
        endpoint = endpoint_dict[source.endpoint]

        records = endpoint.load_fixes(  # type: ignore
            load_operation, source=source, records=records
        )
        return records


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
"""Create a commit, including the CoLRev report."""
from __future__ import annotations

import importlib
import os
import shutil
import sys
import tempfile
from importlib.metadata import version
from pathlib import Path
from typing import Optional

import git

import colrev.env.utils
import colrev.exceptions as colrev_exceptions

if False:  # pylint: disable=using-constant-test
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        import colrev.review_manager


class Commit:
    """Create commits"""

    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    __temp_path = Path.home().joinpath("colrev") / Path(".colrev_temp")

    def __init__(
        self,
        *,
        review_manager: colrev.review_manager.ReviewManager,
        msg: str,
        manual_author: bool,
        script_name: str,
        saved_args: Optional[dict] = None,
    ) -> None:
        self.review_manager = review_manager
        self.manual_author = manual_author

        self.msg = msg
        self.script_name = self.__parse_script_name(script_name=script_name)
        self.saved_args = self.__parse_saved_args(saved_args=saved_args)
        self.tree_hash = ""

        self.last_commit_sha = ""
        try:
            self.last_commit_sha = review_manager.dataset.get_last_commit_sha()
        except ValueError:
            pass

        self.records_committed = review_manager.dataset.records_file.is_file()
        self.completeness_condition = review_manager.get_completeness_condition()
        self.colrev_version = f'version {version("colrev")}'
        sys_v = sys.version
        self.python_version = f'version {sys_v[: sys_v.find(" ")]}'
        stream = os.popen("git --version")
        self.git_version = stream.read().replace("git ", "").replace("\n", "")
        stream = os.popen("docker --version")
        self.docker_version = stream.read().replace("Docker ", "").replace("\n", "")
        if "" == self.docker_version:
            self.docker_version = "Not installed"

        self.ext_script_name = ""
        self.ext_script_version = ""

        if script_name != "":
            ext_script = script_name.split(" ")[0]
            if ext_script != "colrev":
                try:
                    script_version = version(ext_script)
                    self.ext_script_name = script_name
                    self.ext_script_version = f"version {script_version}"
                except importlib.metadata.PackageNotFoundError:
                    pass
        self.__temp_path.mkdir(exist_ok=True, parents=True)

    def __parse_saved_args(self, *, saved_args: Optional[dict] = None) -> str:
        saved_args_str = ""
        if saved_args is not None:
            for key, value in saved_args.items():
                if isinstance(value, (bool, float, int, str)):
                    if value == "":
                        saved_args_str = saved_args_str + f"     --{key} \\\n"
                    else:
                        saved_args_str = saved_args_str + f"     --{key}={value} \\\n"
            # Replace the last backslash (for argument chaining across linebreaks)
            saved_args_str = saved_args_str.rstrip(" \\\n")

        return saved_args_str

    def __parse_script_name(self, *, script_name: str) -> str:
        if "MANUAL" == script_name:
            script_name = "Commit created manually or by external script"
        elif " " in script_name:
            script_name = script_name.replace("colrev cli", "colrev").replace(
                "prescreen_cli", "prescreen"
            )
        return script_name

    def __get_version_flag(self) -> str:
        flag = ""
        if "dirty" in version("colrev"):
            flag = "*"
        return flag

    def __get_commit_report_header(self) -> str:
        template = colrev.env.utils.get_template(
            template_path="template/ops/commit_report_header.txt"
        )
        content = template.render(commit_details=self)

        return content

    def __get_commit_report_details(self) -> str:
        template = colrev.env.utils.get_template(
            template_path="template/ops/commit_report_details.txt"
        )
        content = template.render(commit_details=self)

        return content

    def __get_detailed_processing_report(self) -> str:
        processing_report = ""
        if self.review_manager.report_path.is_file():
            # Reformat
            prefixes = [
                "[('change', 'author',",
                "[('change', 'title',",
                "[('change', 'journal',",
                "[('change', 'booktitle',",
            ]

            with tempfile.NamedTemporaryFile(
                dir=self.__temp_path, mode="r+b", delete=False
            ) as temp:
                with open(self.review_manager.report_path, "r+b") as file:
                    shutil.copyfileobj(file, temp)
            # self.report_path.rename(temp.name)
            with open(temp.name, encoding="utf8") as reader, open(
                self.review_manager.report_path, "w", encoding="utf8"
            ) as writer:
                line = reader.readline()
                while line:
                    if (
                        any(prefix in line for prefix in prefixes)
                        and "', '" in line[30:]
                    ):
                        split_pos = line.rfind("', '") + 2
                        indent = line.find("', (") + 3
                        writer.write(line[:split_pos] + "\n")
                        writer.write(" " * indent + line[split_pos:])
                    else:
                        writer.write(line)

                    line = reader.readline()

            with open(self.review_manager.report_path, encoding="utf8") as file:
                line = file.readline()
                debug_part = False
                while line:
                    # For more efficient debugging (loading of dict with Enum)
                    if "colrev_status" in line and "<RecordState." in line:
                        line = line.replace("<RecordState", "RecordState")
                        line = line[: line.rfind(":")] + line[line.rfind(">") + 1 :]
                    if "[DEBUG]" in line or debug_part:
                        debug_part = True
                        if any(
                            x in line
                            for x in ["[INFO]", "[ERROR]", "[WARNING]", "[CRITICAL"]
                        ):
                            debug_part = False
                    if not debug_part:
                        processing_report = processing_report + line
                    line = file.readline()

            processing_report = "\nProcessing report\n" + "".join(processing_report)
        return processing_report

    def __get_commit_report(self) -> str:
        status_operation = self.review_manager.get_status_operation()

        report = self.__get_commit_report_header()
        report += status_operation.get_review_status_report(colors=False)
        report += self.__get_commit_report_details()

        return report

    def create(self) -> bool:
        """Create a commit (including the commit message and details)"""

        if self.review_manager.dataset.has_changes():
            self.review_manager.logger.debug("Prepare commit: checks and updates")
            self.review_manager.update_status_yaml()
            self.review_manager.dataset.add_changes(
                path=self.review_manager.STATUS_RELATIVE
            )

            committer, email = self.review_manager.get_committer()

            if self.manual_author:
                git_author = git.Actor(committer, email)
            else:
                git_author = git.Actor(f"script:{self.script_name}", email)

            # Note : this should run as the last command before creating the commit
            # to ensure that the git tree_hash is up-to-date.
            self.tree_hash = self.review_manager.dataset.get_tree_hash()
            self.msg = (
                self.msg
                + self.__get_version_flag()
                + self.__get_commit_report()
                + self.__get_detailed_processing_report()
            )
            self.review_manager.dataset.create_commit(
                msg=self.msg,
                author=git_author,
                committer=git.Actor(committer, email),
                hook_skipping=True,
            )

            self.review_manager.logger.info("Created commit")
            self.review_manager.reset_report_logger()
            if self.review_manager.dataset.has_changes():
                raise colrev_exceptions.DirtyRepoAfterProcessingError(
                    "A clean repository is expected."
                )
            return True

        return False

    def update_report(self, *, msg_file: Path) -> None:
        """Update the report"""
        report = self.__get_commit_report()
        with open(msg_file, "a", encoding="utf8") as file:
            file.write(report)


if __name__ == "__main__":
    pass



#! /usr/bin/env python
"""SearchSource: AIS electronic Library"""
from __future__ import annotations

import json
import re
import typing
import urllib.parse
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import urlparse

import requests
import zope.interface
from dacite import from_dict
from dataclasses_jsonschema import JsonSchemaMixin

import colrev.env.package_manager
import colrev.exceptions as colrev_exceptions
import colrev.ops.search
import colrev.record
import colrev.ui_cli.cli_colors as colors

# pylint: disable=unused-argument
# pylint: disable=duplicate-code


@zope.interface.implementer(
    colrev.env.package_manager.SearchSourcePackageEndpointInterface
)
@dataclass
class AISeLibrarySearchSource(JsonSchemaMixin):
    """SearchSource for the AIS electronic Library (AISeL)"""

    settings_class = colrev.env.package_manager.DefaultSourceSettings
    source_identifier = "url"
    search_type = colrev.settings.SearchType.DB
    api_search_supported = True
    ci_supported: bool = True
    heuristic_status = colrev.env.package_manager.SearchSourceHeuristicStatus.supported
    short_name = "AIS eLibrary"
    link = (
        "https://github.com/CoLRev-Environment/colrev/blob/main/"
        + "colrev/ops/built_in/search_sources/aisel.md"
    )

    def __init__(
        self, *, source_operation: colrev.operation.CheckOperation, settings: dict
    ) -> None:
        self.search_source = from_dict(data_class=self.settings_class, data=settings)
        self.review_manager = source_operation.review_manager

    @classmethod
    def heuristic(cls, filename: Path, data: str) -> dict:
        """Source heuristic for AIS electronic Library (AISeL)"""

        result = {"confidence": 0.0}
        # TBD: aisel does not return bibtex?!
        nr_ais_links = data.count("https://aisel.aisnet.org/")
        nr_items = data.count("\n@")
        if nr_items > 0 and "colrev_status" not in data:
            result["confidence"] = nr_ais_links / nr_items

        if "%T " in data and "colrev_status" not in data:
            if data.count("%U https://aisel.aisnet.org") > 0.9 * data.count("%T "):
                result["confidence"] = 1.0

        if "@article" in data or "@inproc" in data and "colrev_status" not in data:
            if data.count("https://aisel.aisnet.org") > 0.9 * data.count("\n@"):
                result["confidence"] = 1.0

        return result

    @classmethod
    def add_endpoint(
        cls, search_operation: colrev.ops.search.Search, query: str
    ) -> typing.Optional[colrev.settings.SearchSource]:
        """Add SearchSource as an endpoint (based on query provided to colrev search -a )"""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-statements

        query = query.lstrip("colrev.ais_library:").rstrip('"').lstrip('"')

        host = urlparse(query).hostname

        if host and host.endswith("aisel.aisnet.org"):
            peer_reviewed = "peer_reviewed=true" in query
            start_date = ""
            if "start_date=" in query:
                start_date = query[query.find("start_date=") + 11 :]
                start_date = start_date[: start_date.find("&")]
                start_date = start_date.replace("%2F", "/")
            end_date = ""
            if "end_date=" in query:
                end_date = query[query.find("end_date=") + 9 :]
                end_date = end_date[: end_date.find("&")]
                end_date = end_date.replace("%2F", "/")

            query = query[query.find("?q=") + 3 : query.find("&start")]
            query_parts = urllib.parse.unquote(query).split(" ")

            search_terms = []
            query_parts_merged = []
            parenthesis_expression = ""
            for query_part in query_parts:
                if query_part not in ["(", ")"] and "" == parenthesis_expression:
                    query_parts_merged.append(query_part)
                elif "(" == query_part:
                    parenthesis_expression += "("
                elif ")" == query_part:
                    parenthesis_expression = parenthesis_expression.rstrip().replace(
                        "(", ""
                    )
                    query_parts_merged.append(parenthesis_expression)
                    parenthesis_expression = ""
                else:
                    parenthesis_expression = parenthesis_expression + query_part + " "

            term_no = 1
            operator = ""

            file_query = ""
            for query_part in query_parts_merged:
                if query_part in ["OR", "AND", "NOT"]:
                    operator = query_part
                    file_query += "_" + query_part + "_"
                    continue

                field = "All fields"
                if "%3A" in query_part:
                    field, query_part = query_part.split("%3A")
                search_term = {"operator": operator, "term": query_part, "field": field}
                file_query += "_" + query_part + "_"

                search_terms.append(search_term)
                term_no += 1

            params = {"search_terms": search_terms, "peer_reviewed": peer_reviewed}
            if start_date:
                params["start_date"] = start_date
            if end_date:
                params["end_date"] = end_date

            file_query = "aisel_" + file_query.lstrip("_").rstrip("_").replace(
                "__", "_"
            ).replace("%22", "").replace("*", "")

            filename = search_operation.get_unique_filename(
                file_path_string=f"ais_{file_query}"
            )
            add_source = colrev.settings.SearchSource(
                endpoint="colrev.ais_library",
                filename=filename,
                search_type=colrev.settings.SearchType.DB,
                search_parameters={"query": params},
                load_conversion_package_endpoint={"endpoint": "colrev.bibtex"},
                comment="",
            )
            return add_source

        return None

    def validate_source(
        self,
        search_operation: colrev.ops.search.Search,
        source: colrev.settings.SearchSource,
    ) -> None:
        """Validate the SearchSource (parameters etc.)"""

        search_operation.review_manager.logger.debug(
            f"Validate SearchSource {source.filename}"
        )

        if "query" in source.search_parameters:
            if "search_terms" not in source.search_parameters["query"]:
                raise colrev_exceptions.InvalidQueryException(
                    "query parameter does not contain search_terms"
                )

        # Note : can simply add files downloaded from AIS
        # raise colrev_exceptions.InvalidQueryException(
        #     f"Source missing query_file or query search_parameter ({source.filename})"
        # )

        if "query_file" in source.search_parameters:
            if not Path(source.search_parameters["query_file"]).is_file():
                raise colrev_exceptions.InvalidQueryException(
                    f"File does not exist: query_file {source.search_parameters['query_file']} "
                    f"for ({source.filename})"
                )

        search_operation.review_manager.logger.debug(
            f"SearchSource {source.filename} validated"
        )

    def __get_ais_query_return(self) -> list:
        def query_from_params(params: dict) -> str:
            final = ""
            for i in params["search_terms"]:
                final = f"{final} {i['operator']} {i['field']}:{i['term']}".strip()
                final = final.replace("All fields:", "")

            if "start_date" in params["search_terms"]:
                final = f"{final}&start_date={params['search_terms']['start_date']}"

            if "end_date" in params["search_terms"]:
                final = f"{final}&end_date={params['search_terms']['end_date']}"

            if "peer_reviewed" in params["search_terms"]:
                final = f"{final}&peer_reviewed=true"

            return urllib.parse.quote(final)

        params = self.search_source.search_parameters["query"]
        final_q = query_from_params(params)

        query_string = (
            "https://aisel.aisnet.org/do/search/results/refer?"
            + "start=0&context=509156&sort=score&facet=&dlt=Export122204"
        )
        query_string = f"{query_string}&q={final_q}"

        response = requests.get(query_string)
        response.raise_for_status()

        zotero_translation_service = (
            self.review_manager.get_zotero_translation_service()
        )
        zotero_translation_service.start()

        headers = {"Content-type": "text/plain"}
        ret = requests.post(
            "http://127.0.0.1:1969/import",
            headers=headers,
            data=response.content,
            timeout=30,
        )
        headers = {"Content-type": "application/json"}

        try:
            json_content = json.loads(ret.content)
            export = requests.post(
                "http://127.0.0.1:1969/export?format=bibtex",
                headers=headers,
                json=json_content,
                timeout=30,
            )

        except Exception as exc:
            raise colrev_exceptions.ImportException(
                f"Zotero translators failed ({exc})"
            )

        records = self.review_manager.dataset.load_records_dict(
            load_str=export.content.decode("utf-8")
        )

        return list(records.values())

    def __run_parameter_search(
        self,
        *,
        search_operation: colrev.ops.search.Search,
        ais_feed: colrev.ops.search.GeneralOriginFeed,
        rerun: bool,
    ) -> None:
        # pylint: disable=too-many-branches

        if rerun:
            search_operation.review_manager.logger.info(
                "Performing a search of the full history (may take time)"
            )

        records = search_operation.review_manager.dataset.load_records_dict()
        nr_retrieved, nr_changed = 0, 0

        try:
            for record_dict in self.__get_ais_query_return():
                # Note : discard "empty" records
                if "" == record_dict.get("author", "") and "" == record_dict.get(
                    "title", ""
                ):
                    continue

                try:
                    ais_feed.set_id(record_dict=record_dict)
                except colrev_exceptions.NotFeedIdentifiableException:
                    continue

                # prev_record_dict_version = {}
                # if record_dict["ID"] in ais_feed.feed_records:
                #     prev_record_dict_version = deepcopy(
                #         ais_feed.feed_records[record_dict["ID"]]
                #     )

                prep_record = colrev.record.PrepRecord(data=record_dict)

                if "colrev_data_provenance" in prep_record.data:
                    del prep_record.data["colrev_data_provenance"]

                added = ais_feed.add_record(record=prep_record)

                if added:
                    search_operation.review_manager.logger.info(
                        " retrieve " + prep_record.data["url"]
                    )
                    nr_retrieved += 1
                # else:
                #     changed = search_operation.update_existing_record(
                #         records=records,
                #         record_dict=prep_record.data,
                #         prev_record_dict_version=prev_record_dict_version,
                #         source=self.search_source,
                #         update_time_variant_fields=rerun,
                #     )
                #     if changed:
                #         nr_changed += 1

            if nr_retrieved > 0:
                search_operation.review_manager.logger.info(
                    f"{colors.GREEN}Retrieved {nr_retrieved} records{colors.END}"
                )
            else:
                search_operation.review_manager.logger.info(
                    f"{colors.GREEN}No additional records retrieved{colors.END}"
                )

            if nr_changed > 0:
                self.review_manager.logger.info(
                    f"{colors.GREEN}Updated {nr_changed} records{colors.END}"
                )
            else:
                if records:
                    self.review_manager.logger.info(
                        f"{colors.GREEN}Records (data/records.bib) up-to-date{colors.END}"
                    )

            ais_feed.save_feed_file()
            search_operation.review_manager.dataset.save_records_dict(records=records)
            search_operation.review_manager.dataset.add_record_changes()

        except requests.exceptions.JSONDecodeError as exc:
            # watch github issue:
            # https://github.com/fabiobatalha/crossrefapi/issues/46
            if "504 Gateway Time-out" in str(exc):
                raise colrev_exceptions.ServiceNotAvailableException(
                    "Crossref (check https://status.crossref.org/)"
                )
            raise colrev_exceptions.ServiceNotAvailableException(
                f"Crossref (check https://status.crossref.org/) ({exc})"
            )

    def run_search(
        self, search_operation: colrev.ops.search.Search, rerun: bool
    ) -> None:
        """Run a search of AISeLibrary"""

        ais_feed = self.search_source.get_feed(
            review_manager=search_operation.review_manager,
            source_identifier=self.source_identifier,
            update_only=(not rerun),
        )

        #  Note: not (yet) supported: self.search_source.is_md_source()
        if "query" in self.search_source.search_parameters:
            self.__run_parameter_search(
                search_operation=search_operation,
                ais_feed=ais_feed,
                rerun=rerun,
            )

    def get_masterdata(
        self,
        prep_operation: colrev.ops.prep.Prep,
        record: colrev.record.Record,
        save_feed: bool = True,
        timeout: int = 10,
    ) -> colrev.record.Record:
        """Not implemented"""
        return record

    def load_fixes(
        self,
        load_operation: colrev.ops.load.Load,
        source: colrev.settings.SearchSource,
        records: typing.Dict,
    ) -> dict:
        """Load fixes for AIS electronic Library (AISeL)"""

        return records

    def prepare(
        self, record: colrev.record.PrepRecord, source: colrev.settings.SearchSource
    ) -> colrev.record.Record:
        """Source-specific preparation for the AIS electronic Library (AISeL)"""

        # pylint: disable=too-many-branches
        # pylint: disable=too-many-statements

        ais_mapping: dict = {}
        record.rename_fields_based_on_mapping(mapping=ais_mapping)

        # Note : simple heuristic
        # but at the moment, AISeLibrary only indexes articles and conference papers
        if (
            record.data.get("volume", "UNKNOWN") != "UNKNOWN"
            or record.data.get("number", "UNKNOWN") != "UNKNOWN"
        ) and not any(
            x in record.data.get("journal", "")
            for x in [
                "HICSS",
                "ICIS",
                "ECIS",
                "AMCIS",
                "Proceedings",
                "All Sprouts Content",
            ]
        ):
            if (
                "journal" not in record.data
                and "title" in record.data
                and "chapter" in record.data
            ):
                record.rename_field(key="title", new_key="journal")
                record.rename_field(key="chapter", new_key="title")
                record.remove_field(key="publisher")

            record.change_entrytype(new_entrytype="article")

        else:
            record.remove_field(key="publisher")

            if (
                "booktitle" not in record.data
                and "title" in record.data
                and "chapter" in record.data
            ):
                record.rename_field(key="title", new_key="booktitle")
                record.rename_field(key="chapter", new_key="title")

            record.change_entrytype(new_entrytype="inproceedings")

            if record.data.get("booktitle", "") in [
                "Research-in-Progress Papers",
                "Research Papers",
            ]:
                if "https://aisel.aisnet.org/ecis" in record.data.get("url", ""):
                    record.update_field(
                        key="booktitle", value="ECIS", source="prep_ais_source"
                    )

        if record.data.get("journal", "") == "Management Information Systems Quarterly":
            record.update_field(
                key="journal", value="MIS Quarterly", source="prep_ais_source"
            )

        if "inproceedings" == record.data["ENTRYTYPE"]:
            if "ICIS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="International Conference on Information Systems",
                    source="prep_ais_source",
                )
            if "PACIS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="Pacific-Asia Conference on Information Systems",
                    source="prep_ais_source",
                )
            if "ECIS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="European Conference on Information Systems",
                    source="prep_ais_source",
                )
            if "AMCIS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="Americas Conference on Information Systems",
                    source="prep_ais_source",
                )
            if "HICSS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="Hawaii International Conference on System Sciences",
                    source="prep_ais_source",
                )
            if "MCIS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="Mediterranean Conference on Information Systems",
                    source="prep_ais_source",
                )
            if "ACIS" in record.data.get("booktitle", ""):
                record.update_field(
                    key="booktitle",
                    value="Australasian Conference on Information Systems",
                    source="prep_ais_source",
                )
        if "https://aisel.aisnet.org/hicss" in record.data.get("url", ""):
            record.update_field(
                key="booktitle",
                value="Hawaii International Conference on System Sciences",
                source="prep_ais_source",
            )
        elif "https://aisel.aisnet.org/amcis" in record.data.get("url", ""):
            record.update_field(
                key="booktitle",
                value="Americas Conference on Information Systems",
                source="prep_ais_source",
            )

        elif "https://aisel.aisnet.org/pacis" in record.data.get("url", ""):
            record.update_field(
                key="booktitle",
                value="Pacific-Asia Conference on Information Systems",
                source="prep_ais_source",
            )
        elif "https://aisel.aisnet.org/ecis" in record.data.get("url", ""):
            record.update_field(
                key="booktitle",
                value="European Conference on Information Systems",
                source="prep_ais_source",
            )
        elif "https://aisel.aisnet.org/icis" in record.data.get("url", ""):
            record.update_field(
                key="booktitle",
                value="International Conference on Information Systems",
                source="prep_ais_source",
            )
        elif "https://aisel.aisnet.org/bise/" in record.data.get("url", ""):
            record.update_field(
                key="journal",
                value="Business & Information Systems Engineering",
                source="prep_ais_source",
            )

        if "abstract" in record.data:
            if "N/A" == record.data["abstract"]:
                record.remove_field(key="abstract")
        if "author" in record.data:
            record.update_field(
                key="author",
                value=record.data["author"].replace("\n", " "),
                source="prep_ais_source",
                keep_source_if_equal=True,
            )

        if re.match(
            r"MISQ Volume \d{1,2}, Issue \d Table of Contents",
            record.data.get("title", ""),
        ):
            record.prescreen_exclude(reason="complementary material")

        return record


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
"""Web-UI editor for CoLRev project settings"""
from __future__ import annotations

import json
import os
import signal
import webbrowser
from pathlib import Path
from threading import Timer

from flask import Flask
from flask import jsonify
from flask import request
from flask import Response
from flask import send_from_directory
from flask_cors import CORS

import colrev.env.package_manager
import colrev.review_manager
import colrev.settings


class SettingsEditor:
    """A web-based editor for CoLRev settings"""

    # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods

    settings_path: Path
    review_manager: colrev.review_manager.ReviewManager

    def __init__(self, *, review_manager: colrev.review_manager.ReviewManager) -> None:
        self.review_manager = review_manager
        self.package_manager = self.review_manager.get_package_manager()
        self.settings_path = self.review_manager.settings_path

        # Note : no need for default values (they are already inserted before by the template setup)

    def _open_browser(self) -> None:
        url = "http://127.0.0.1:5000"

        Timer(1, lambda: webbrowser.open_new(url)).start()
        print(f"Open at {url}")

    def open_settings_editor(self) -> None:
        """Open the settings editor"""

        app = Flask(__name__, static_url_path="", static_folder="build")
        CORS(app)

        app.config["path"] = str(self.settings_path)

        @app.route("/", defaults={"path": ""})
        def serve(path: str) -> Response:  # pylint: disable=unused-argument
            assert app.static_folder
            return send_from_directory(Path(app.static_folder), "index.html")

        @app.route("/<path:filename>")
        def base_static(filename: str) -> Response:
            return send_from_directory(app.root_path + "/", filename)

        @app.route("/api/getSettings")
        def getSettings() -> Response:
            with open(self.settings_path, encoding="utf-8") as file:
                json__content = file.read()

            response = app.response_class(
                response=json__content, mimetype="application/json"
            )

            return response

        @app.route("/api/saveSettings", methods=["POST"])
        def saveSettings() -> str:
            commit_selected = request.args.get("commitSelected")

            with open(self.settings_path, "w", encoding="utf-8") as outfile:
                json_string = json.dumps(request.json, indent=4)
                outfile.write(json_string)
            if commit_selected:
                self.review_manager.dataset.add_changes(path=self.settings_path)
                self.review_manager.create_commit(msg="Update settings")

            return "ok"

        @app.route("/api/getOptions")
        def getOptions() -> Response:
            # Decision: get the whole list of setting_options (not individually)
            # "similarity": {'type': 'float', 'min': 0, 'max': 1}
            options = self.review_manager.settings.get_settings_schema()

            return jsonify(options)

        @app.route("/api/getPackages")
        def getPackages() -> Response:
            package_type_string = request.args.get("PackageEndpointType")

            package_type = colrev.env.package_manager.PackageEndpointType[
                package_type_string
            ]
            discovered_packages = self.package_manager.discover_packages(
                package_type=package_type
            )

            return jsonify(discovered_packages)

        # pylint: disable=unused-argument
        @app.route("/api/getPackageDetails")
        def getPackageDetails() -> Response:
            package_type_string = request.args.get("PackageEndpointType")
            package_identifier = request.args.get("PackageIdentifier")
            # endpoint_version = request.args.get("EndpointVersion")

            package_type = colrev.env.package_manager.PackageEndpointType[
                package_type_string
            ]
            package_details = self.package_manager.get_package_details(
                package_type=package_type, package_identifier=package_identifier
            )

            return jsonify(package_details)

        @app.get("/api/shutdown")
        def shutdown() -> str:
            timer = Timer(3.0, terminate_process)
            timer.start()  # after 3 seconds
            return "ok"

        def terminate_process() -> None:
            pid = os.getpid()
            os.kill(pid, signal.SIGTERM)

        self._open_browser()
        app.run(host="0.0.0.0", port=5000, use_reloader=False)


def main() -> None:
    """Main entrypoint for the settings editor"""

    review_manager = colrev.review_manager.ReviewManager()
    se_instance = SettingsEditor(review_manager=review_manager)
    se_instance.open_settings_editor()


if __name__ == "__main__":
    main()
